# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2013, Mathieu Duponchelle <mduponchelle1@gmail.com>
# Copyright (c) 2016, Thibault Saunier <tsaunier@gnome.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import os
from gettext import gettext as _

import numpy
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GES
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstController
from gi.repository import Gtk
from matplotlib.axes import Axes
from matplotlib.backend_bases import MouseButton
from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo
from matplotlib.collections import PathCollection
from matplotlib.figure import Figure
from matplotlib.lines import Line2D

from pitivi.configure import get_pixmap_dir
from pitivi.effects import ALLOWED_ONLY_ONCE_EFFECTS
from pitivi.timeline.previewers import AudioPreviewer
from pitivi.timeline.previewers import ImagePreviewer
from pitivi.timeline.previewers import TitlePreviewer
from pitivi.timeline.previewers import VideoPreviewer
from pitivi.undo.timeline import CommitTimelineFinalizingAction
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import disconnect_all_by_func
from pitivi.utils.misc import filename_from_uri
from pitivi.utils.pipeline import PipelineError
from pitivi.utils.timeline import SELECT
from pitivi.utils.timeline import SELECT_ADD
from pitivi.utils.timeline import Selected
from pitivi.utils.timeline import UNSELECT
from pitivi.utils.timeline import Zoomable
from pitivi.utils.ui import EFFECT_TARGET_ENTRY
from pitivi.utils.ui import set_children_state_recurse
from pitivi.utils.ui import unset_children_state_recurse

KEYFRAME_LINE_HEIGHT = 2
KEYFRAME_LINE_ALPHA = 0.5
KEYFRAME_LINE_COLOR = "#EDD400"  # "Tango" medium yellow
KEYFRAME_NODE_COLOR = "#F57900"  # "Tango" medium orange
SELECTED_KEYFRAME_NODE_COLOR = "#204A87"  # "Tango" dark sky blue
HOVERED_KEYFRAME_NODE_COLOR = "#3465A4"  # "Tango" medium sky blue

CURSORS = {
    GES.Edge.EDGE_START: Gdk.Cursor.new(Gdk.CursorType.LEFT_SIDE),
    GES.Edge.EDGE_END: Gdk.Cursor.new(Gdk.CursorType.RIGHT_SIDE)
}

NORMAL_CURSOR = Gdk.Cursor.new(Gdk.CursorType.LEFT_PTR)
DRAG_CURSOR = Gdk.Cursor.new(Gdk.CursorType.HAND1)


def get_pspec(element_factory_name, propname):
    element = Gst.ElementFactory.make(element_factory_name)
    if not element:
        return None

    return [prop for prop in element.list_properties() if prop.name == propname][0]


class KeyframeCurve(FigureCanvasGTK3Cairo, Loggable):
    YLIM_OVERRIDES = {}

    __YLIM_OVERRIDES_VALUES = [("volume", "volume", (0.0, 0.2))]

    for factory_name, propname, values in __YLIM_OVERRIDES_VALUES:
        pspec = get_pspec(factory_name, propname)
        if pspec:
            YLIM_OVERRIDES[pspec] = values

    __gsignals__ = {
        # Signal the keyframes or the curve are being hovered
        "enter": (GObject.SignalFlags.RUN_LAST, None, ()),
        # Signal the keyframes or the curve are not being hovered anymore
        "leave": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, timeline, binding, ges_elem):
        figure = Figure()
        FigureCanvasGTK3Cairo.__init__(self, figure)
        Loggable.__init__(self)

        self._ges_elem = ges_elem
        self._timeline = timeline
        self.__source = binding.props.control_source
        self._connect_sources()
        self.__property_name = binding.props.name
        self.__paramspec = binding.pspec
        self.get_style_context().add_class("KeyframeCurve")

        self.__ylim_min, self.__ylim_max = KeyframeCurve.YLIM_OVERRIDES.get(
            binding.pspec, (0.0, 1.0))
        self.__ydata_drag_start = self.__ylim_min

        # Curve values, basically separating source.get_values() timestamps
        # and values.
        self._line_xs = []
        self._line_ys = []

        # facecolor to None for transparency
        self._ax: Axes = figure.add_axes([0, 0, 1, 1], facecolor='None')
        # Clear the Axes object.
        self._ax.cla()
        self._ax.grid(False)
        self._ax.tick_params(axis='both',
                             which='both',
                             bottom=False,
                             top=False,
                             right=False,
                             left=False)

        # This seems to also be necessary for transparency ..
        figure.patch.set_visible(False)

        # The PathCollection object holding the keyframes dots.
        sizes = [50]
        self._keyframes: PathCollection = self._ax.scatter([], [], marker='D', s=sizes,
                                                           c=KEYFRAME_NODE_COLOR, zorder=2)

        # matplotlib weirdness, simply here to avoid a warning ..
        self._keyframes.set_picker(True)

        # The Line2D object holding the lines between keyframes.
        self.__line: Line2D = self._ax.plot([], [],
                                            alpha=KEYFRAME_LINE_ALPHA,
                                            c=KEYFRAME_LINE_COLOR,
                                            linewidth=KEYFRAME_LINE_HEIGHT, zorder=1)[0]
        self._update_plots()

        # Drag and drop logic
        # Whether the clicked keyframe or line has been dragged.
        self._dragged = False
        # The inpoint of the clicked keyframe.
        self._offset = None
        # The initial keyframe value when a keyframe is being moved.
        self._initial_value = 0
        # The initial keyframe timestamp when a keyframe is being moved.
        self._initial_timestamp = 0
        # The initial event.x when a keyframe is being moved.
        self._initial_x = 0
        # The initial event.y when a keyframe is being moved.
        self._initial_y = 0
        # The (offset, value) of both keyframes of the clicked keyframe line.
        self.__clicked_line = ()
        # Whether the mouse events go to the keyframes logic.
        self.handling_motion = False

        self.__hovered = False

        self.connect("motion-notify-event", self.__motion_notify_event_cb)
        self.connect("event", self._event_cb)
        self.connect("notify::height-request", self.__height_request_cb)
        self.connect("button_release_event", self._button_release_event_cb)

        self.mpl_connect('button_press_event', self._mpl_button_press_event_cb)
        self.mpl_connect('button_release_event', self._mpl_button_release_event_cb)
        self.mpl_connect('motion_notify_event', self._mpl_motion_event_cb)

    def release(self):
        disconnect_all_by_func(self, self.__height_request_cb)
        disconnect_all_by_func(self, self.__motion_notify_event_cb)
        disconnect_all_by_func(self, self._button_release_event_cb)
        disconnect_all_by_func(self, self._control_source_changed_cb)

    def _connect_sources(self):
        self.__source.connect("value-added", self._control_source_changed_cb)
        self.__source.connect("value-removed", self._control_source_changed_cb)
        self.__source.connect("value-changed", self._control_source_changed_cb)

    def _update_plots(self):
        values = self.__source.get_all()
        if len(values) < 2:
            # No plot for less than two points.
            return

        self._line_xs = []
        self._line_ys = []
        for value in values:
            self._line_xs.append(value.timestamp)
            self._line_ys.append(value.value)

        self._populate_lines()

    def _populate_lines(self):
        self._ax.set_xlim(self._line_xs[0], self._line_xs[-1])
        self.__compute_ylim()

        arr = numpy.array((self._line_xs, self._line_ys)).transpose()
        self._keyframes.set_offsets(arr)
        self.__line.set_xdata(self._line_xs)
        self.__line.set_ydata(self._line_ys)
        self.queue_draw()

    def __compute_ylim(self):
        height = self.props.height_request
        if height <= 0:
            return

        ylim_min = -(KEYFRAME_LINE_HEIGHT / height)
        ylim_max = (self.__ylim_max * height) / (height - KEYFRAME_LINE_HEIGHT)
        self._ax.set_ylim(ylim_min, ylim_max)

    def __height_request_cb(self, unused_self, unused_pspec):
        self.__compute_ylim()

    def __maybe_create_keyframe(self, event):
        line_contains = self.__line.contains(event)[0]
        keyframe_existed = self._keyframes.contains(event)[0]
        if line_contains and not keyframe_existed:
            self._create_keyframe(event.xdata)

    def _create_keyframe(self, timestamp):
        res, value = self.__source.control_source_get_value(timestamp)
        assert res
        self.debug("Create keyframe at (%lf, %lf)", timestamp, value)
        with self._timeline.app.action_log.started("Keyframe added",
                                                   toplevel=True):
            self.__source.set(timestamp, value)

    def _remove_keyframe(self, timestamp):
        self.debug("Removing keyframe at timestamp %lf", timestamp)
        with self._timeline.app.action_log.started("Remove keyframe",
                                                   toplevel=True):
            self.__source.unset(timestamp)

    def _move_keyframe(self, source_timestamp, dest_timestamp, dest_value):
        self.__source.unset(source_timestamp)
        self.__source.set(dest_timestamp, dest_value)

    def _move_keyframe_line(self, line, y_dest_value, y_start_value):
        delta = y_dest_value - y_start_value
        for offset, value in line:
            value = max(self.__ylim_min, min(value + delta, self.__ylim_max))
            self.__source.set(offset, value)

    def toggle_keyframe(self, offset):
        """Sets or unsets the keyframe at the specified offset."""
        items = self.__source.get_all()
        if offset in (items[0].timestamp, items[-1].timestamp):
            return

        if offset in [item.timestamp for item in items]:
            self.__source.unset(offset)
        else:
            res, value = self.__source.control_source_get_value(offset)
            assert res
            self.__source.set(offset, value)

    def _control_source_changed_cb(self, unused_control_source, unused_timed_value):
        self._update_plots()
        self._timeline.ges_timeline.get_parent().commit_timeline()

    def __motion_notify_event_cb(self, unused_widget, unused_event):
        # We need to do this here, because Matplotlib's callbacks can't stop
        # signal propagation.
        if self.handling_motion:
            return True
        return False

    def _event_cb(self, unused_element, event):
        if event.type == Gdk.EventType.LEAVE_NOTIFY:
            cursor = NORMAL_CURSOR
            self._timeline.get_window().set_cursor(cursor)
        return False

    def _mpl_button_press_event_cb(self, event):
        if event.button != MouseButton.LEFT:
            return

        result = self._keyframes.contains(event)
        if result[0]:
            # A keyframe has been clicked.
            keyframe_index = result[1]['ind'][0]
            offsets = self._keyframes.get_offsets()
            offset, value = offsets[keyframe_index]

            # pylint: disable=protected-access
            if event.guiEvent.type == Gdk.EventType._2BUTTON_PRESS:
                index = result[1]['ind'][0]
                # pylint: disable=consider-using-in
                if index == 0 or index == len(offsets) - 1:
                    # It's an edge keyframe. These should not be removed.
                    return

                # Rollback the last operation if it is "Move keyframe".
                # This is needed because a double-click also triggers a
                # BUTTON_PRESS event which starts a "Move keyframe" operation
                self._timeline.app.action_log.try_rollback("Move keyframe")
                self._offset = None

                # A keyframe has been double-clicked, remove it.
                self._remove_keyframe(offset)
            else:
                # Remember the clicked frame for drag&drop.
                self._timeline.app.action_log.begin("Move keyframe",
                                                    toplevel=True)
                self._initial_x = event.x
                self._initial_y = event.y
                self._offset = offset
                self._initial_timestamp = offset
                self._initial_value = value
                self.handling_motion = True
            return

        result = self.__line.contains(event)
        if result[0]:
            # The line has been clicked.
            self.debug("The keyframe curve has been clicked")
            self._timeline.app.action_log.begin("Move keyframe curve segment",
                                                toplevel=True)
            x = event.xdata
            offsets = self._keyframes.get_offsets()
            keyframes = offsets[:, 0]
            right = numpy.searchsorted(keyframes, x)
            # Remember the clicked line for drag&drop.
            self.__clicked_line = (offsets[right - 1], offsets[right])
            self.__ydata_drag_start = max(self.__ylim_min, min(event.ydata, self.__ylim_max))
            self.handling_motion = True

    def _mpl_motion_event_cb(self, event):
        if event.ydata is not None and event.xdata is not None:
            # The mouse event is in the figure boundaries.
            if self._offset is not None:
                self._dragged = True
                keyframe_ts, ydata = self.__compute_keyframe_position(event)
                self._move_keyframe(int(self._offset), keyframe_ts, ydata)
                self._offset = keyframe_ts
                self._update_tooltip(event)
                hovering = True
            elif self.__clicked_line:
                self._dragged = True
                ydata = max(self.__ylim_min, min(event.ydata, self.__ylim_max))
                self._move_keyframe_line(self.__clicked_line, ydata, self.__ydata_drag_start)
                hovering = True
            else:
                hovering = self.__line.contains(event)[0]
        else:
            hovering = False

        if hovering:
            cursor = DRAG_CURSOR
            self._update_tooltip(event)
            if not self.__hovered:
                self.emit("enter")
                self.__hovered = True
        else:
            cursor = NORMAL_CURSOR
            if self.__hovered:
                self.emit("leave")
                self._update_tooltip(None)
                self.__hovered = False

        self._timeline.get_window().set_cursor(cursor)

    def _mpl_button_release_event_cb(self, event):
        if event.button != MouseButton.LEFT:
            return

        # In order to make sure we seek to the exact position where we added a
        # new keyframe, we don't use matplotlib's event.xdata, but rather
        # compute it the same way we do for the seek logic.
        event_widget = Gtk.get_event_widget(event.guiEvent)
        x, unused_y = event_widget.translate_coordinates(self._timeline.layout.layers_vbox,
                                                         event.x, event.y)
        event.xdata = Zoomable.pixel_to_ns(x) - self._ges_elem.props.start + self._ges_elem.props.in_point

        if self._offset is not None:
            # If dragging a keyframe, make sure the keyframe ends up exactly
            # where the mouse was released. Otherwise, the playhead will not
            # seek exactly on the keyframe.
            if self._dragged:
                if event.ydata is not None:
                    keyframe_ts, ydata = self.__compute_keyframe_position(event)
                    self._move_keyframe(int(self._offset), keyframe_ts, ydata)
            self.debug("Keyframe released")
            self._timeline.app.action_log.commit("Move keyframe")
        elif self.__clicked_line:
            self.debug("Line released")
            self._timeline.app.action_log.commit("Move keyframe curve segment")

            if not self._dragged:
                # The keyframe line was clicked, but not dragged
                assert event.guiEvent.type == Gdk.EventType.BUTTON_RELEASE
                self.__maybe_create_keyframe(event)

        self.handling_motion = False
        self._offset = None
        self.__clicked_line = ()

    def _button_release_event_cb(self, unused_widget, event):
        if not event.get_button() == (True, 1):
            return False

        dragged = self._dragged
        self._dragged = False

        # Return True to stop signal propagation, otherwise the clip will be
        # unselected.
        return dragged

    def _update_tooltip(self, event):
        """Sets or clears the tooltip showing info about the hovered line."""
        markup = None
        if event:
            if not event.xdata:
                return
            if self._offset is not None:
                xdata = self._offset
            else:
                xdata = max(self._line_xs[0], min(event.xdata, self._line_xs[-1]))
            res, value = self.__source.control_source_get_value(xdata)
            assert res
            pmin = self.__paramspec.minimum
            pmax = self.__paramspec.maximum
            value = value * (pmax - pmin) + pmin
            # Translators: This is a tooltip for a clip's keyframe curve,
            # showing what the keyframe curve affects, the timestamp at
            # the mouse cursor location, and the value at that timestamp.
            markup = _("Property: %s\nTimestamp: %s\nValue: %s") % (
                self.__property_name,
                Gst.TIME_ARGS(xdata),
                "{:.3f}".format(value))
        self.set_tooltip_markup(markup)

    def __compute_keyframe_position(self, event):
        keyframe_ts = self.__compute_keyframe_new_timestamp(event)
        ydata = max(self.__ylim_min, min(event.ydata, self.__ylim_max))
        if self._timeline.get_parent().control_mask:
            delta_x = abs(event.x - self._initial_x)
            delta_y = abs(event.y - self._initial_y)
            if delta_x > delta_y:
                ydata = self._initial_value
            else:
                keyframe_ts = self._initial_timestamp

        return keyframe_ts, ydata

    def __compute_keyframe_new_timestamp(self, event):
        # The user can not change the timestamp of the first
        # and last keyframes.
        values = self.__source.get_all()
        if self._offset in (values[0].timestamp, values[-1].timestamp):
            return self._offset

        if event.xdata != self._offset:
            try:
                kf = next(kf for kf in values if kf.timestamp == int(self._offset))
            except StopIteration:
                return event.xdata

            i = values.index(kf)
            keyframe_timestamp = int(event.xdata)
            if keyframe_timestamp <= values[i - 1].timestamp:
                keyframe_timestamp = values[i - 1].timestamp + 1
            if keyframe_timestamp >= values[i + 1].timestamp:
                keyframe_timestamp = values[i + 1].timestamp - 1
            return keyframe_timestamp

        return event.xdata


class MultipleKeyframeCurve(KeyframeCurve):
    """Keyframe curve which controls multiple properties at once."""

    def __init__(self, timeline, bindings, ges_elem):
        self.__bindings = bindings
        super().__init__(timeline, bindings[0], ges_elem)

        self._timeline = timeline
        self._project = timeline.app.project_manager.current_project
        self._project.pipeline.connect("position", self._position_cb)

        sizes = [80]
        self.__selected_keyframe = self._ax.scatter([0], [0.5], marker='D', s=sizes,
                                                    c=SELECTED_KEYFRAME_NODE_COLOR, zorder=3)
        self.__hovered_keyframe = self._ax.scatter([0], [0.5], marker='D', s=sizes,
                                                   c=HOVERED_KEYFRAME_NODE_COLOR, zorder=3)
        self.__update_selected_keyframe()
        self.__hovered_keyframe.set_visible(False)

    def release(self):
        super().release()
        self._project.pipeline.disconnect_by_func(self._position_cb)

    def _connect_sources(self):
        for binding in self.__bindings:
            source = binding.props.control_source
            source.connect("value-added", self._control_source_changed_cb)
            source.connect("value-removed", self._control_source_changed_cb)
            source.connect("value-changed", self._control_source_changed_cb)

    def _update_plots(self):
        timestamps = []
        for binding in self.__bindings:
            ts = [value.timestamp for value in binding.props.control_source.get_all()]
            timestamps.extend(ts)
        timestamps = sorted(list(set(timestamps)))

        if len(timestamps) < 2:
            # No plot for less than two points.
            return

        self._line_xs = []
        self._line_ys = []
        for timestamp in timestamps:
            self._line_xs.append(timestamp)
            self._line_ys.append(0.5)

        self._populate_lines()

    def _create_keyframe(self, timestamp):
        with self._timeline.app.action_log.started("Add keyframe",
                                                   toplevel=True):
            for binding in self.__bindings:
                binding.props.control_source.set(timestamp, binding.get_value(timestamp))

    def _remove_keyframe(self, timestamp):
        with self._timeline.app.action_log.started("Remove keyframe",
                                                   toplevel=True):
            for binding in self.__bindings:
                binding.props.control_source.unset(timestamp)

    def _move_keyframe(self, source_timestamp, dest_timestamp, unused_dest_value):
        if source_timestamp == dest_timestamp:
            return

        for binding in self.__bindings:
            dest_value = binding.get_value(source_timestamp)
            binding.props.control_source.set(dest_timestamp, dest_value)
            binding.props.control_source.unset(source_timestamp)

    def _move_keyframe_line(self, line, y_dest_value, y_start_value):
        pass

    def _mpl_button_release_event_cb(self, event):
        if event.button == MouseButton.LEFT:
            if self._offset is not None and not self._dragged:
                # A keyframe was clicked but not dragged, so we
                # should select it by seeking to its position.
                source = self._timeline.selection.get_single_clip()
                assert source
                position = int(self._offset) - source.props.in_point + source.props.start

                if self._timeline.app.settings.leftClickAlsoSeeks:
                    self._timeline.set_next_seek_position(position)
                else:
                    self._project.pipeline.simple_seek(position)

        super()._mpl_button_release_event_cb(event)

    def _mpl_motion_event_cb(self, event):
        super()._mpl_motion_event_cb(event)

        result = self._keyframes.contains(event)
        if result[0]:
            # A keyframe is hovered
            keyframe_index = result[1]['ind'][0]
            offset = self._keyframes.get_offsets()[keyframe_index][0]
            self.__show_special_keyframe(self.__hovered_keyframe, offset)
        else:
            self.__hide_special_keyframe(self.__hovered_keyframe)

    def __show_special_keyframe(self, keyframe, offset):
        offsets = numpy.array([[offset, 0.5]])
        keyframe.set_offsets(offsets)
        keyframe.set_visible(True)
        self.queue_draw()

    def __hide_special_keyframe(self, keyframe):
        keyframe.set_visible(False)
        self.queue_draw()

    def _control_source_changed_cb(self, control_source, timed_value):
        super()._control_source_changed_cb(control_source, timed_value)
        self.__update_selected_keyframe()
        self.__hide_special_keyframe(self.__hovered_keyframe)

    def _position_cb(self, unused_pipeline, unused_position):
        self.__update_selected_keyframe()

    def __update_selected_keyframe(self):
        try:
            position = self._project.pipeline.get_position()
        except PipelineError:
            self.warning("Could not get pipeline position")
            return

        source = self._timeline.selection.get_single_clip()
        if source is None:
            return
        source_position = position - source.props.start + source.props.in_point

        offsets = self._keyframes.get_offsets()
        keyframes = offsets[:, 0]

        index = numpy.searchsorted(keyframes, source_position)
        if 0 <= index < len(keyframes) and keyframes[index] == source_position:
            self.__show_special_keyframe(self.__selected_keyframe, source_position)
        else:
            self.__hide_special_keyframe(self.__selected_keyframe)

    def _update_tooltip(self, event):
        markup = None
        if event:
            if not event.xdata:
                return
            markup = _("Timestamp: %s") % Gst.TIME_ARGS(event.xdata)
        self.set_tooltip_markup(markup)


class TimelineElement(Gtk.Layout, Zoomable, Loggable):
    __gsignals__ = {
        # Signal the keyframes curve are being hovered
        "curve-enter": (GObject.SignalFlags.RUN_LAST, None, ()),
        # Signal the keyframes curve are not being hovered anymore
        "curve-leave": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, element, timeline):
        Gtk.Layout.__init__(self)
        Zoomable.__init__(self)
        Loggable.__init__(self)

        self.set_name(element.get_name())

        self.timeline = timeline
        self._ges_elem = element
        self._ges_elem.selected = Selected()
        self._ges_elem.selected.connect(
            "selected-changed", self.__selected_changed_cb)

        self.__width = 0
        self.__height = 0

        # Needed for effect's keyframe toggling
        self._ges_elem.ui_element = self

        self.props.vexpand = True

        self.__previewer = self._get_previewer()
        if self.__previewer:
            self.add(self.__previewer)

        self.__background = self._get_background()
        if self.__background:
            self.add(self.__background)

        self.keyframe_curve = None
        self.__controlled_property = None
        self.show_all()

        # We set up the default mixing property right here, if a binding was
        # already set (when loading a project), it will be added later
        # and override that one.
        self.show_default_keyframes(lazy_render=True)

    def update_previewer(self):
        """Refreshes the previewer widget."""
        if self.__previewer:
            self.__previewer.refresh()

    def release(self):
        if self.__previewer:
            self.__previewer.release()

    # Public API
    def set_size(self, width, height):
        width = max(0, width)
        self.set_size_request(width, height)

        if self.__previewer:
            self.__previewer.set_size_request(width, height)

        if self.__background:
            self.__background.set_size_request(width, height)

        if self.keyframe_curve:
            self.keyframe_curve.set_size_request(width, height)

        self.__width = width
        self.__height = height

    def show_keyframes(self, ges_elem, prop):
        self.__set_keyframes(ges_elem, prop)
        binding = ges_elem.get_control_binding(prop.name)
        self.__create_keyframe_curve([binding])

    def show_default_keyframes(self, lazy_render=False):
        self.__set_keyframes(self._ges_elem, self._get_default_mixing_property())
        if not lazy_render:
            self.__create_keyframe_curve()

    def show_multiple_keyframes(self, bindings):
        self.__controlled_property = None
        self.__create_keyframe_curve(bindings)

    def __set_keyframes(self, ges_elem, prop):
        self.__remove_keyframes()
        self.__controlled_property = prop
        if self.__controlled_property:
            self.__create_control_binding(ges_elem)

    def __curve_enter_cb(self, unused_keyframe_curve):
        self.emit("curve-enter")

    def __curve_leave_cb(self, unused_keyframe_curve):
        self.emit("curve-leave")

    def __remove_keyframes(self):
        if not self.keyframe_curve:
            # Nothing to remove.
            return

        self.keyframe_curve.disconnect_by_func(self.__curve_enter_cb)
        self.keyframe_curve.disconnect_by_func(self.__curve_leave_cb)
        self.remove(self.keyframe_curve)

        self.keyframe_curve.release()
        self.keyframe_curve = None

    # Private methods
    def __ensure_keyframes(self, binding):
        """Ensures that we have at least 2 keyframes (at inpoint and end)."""
        source = binding.props.control_source
        values = source.get_all()

        if len(values) < 2:
            source.unset_all()
            values_range = self.__controlled_property.maximum - self.__controlled_property.minimum
            val = float(self.__controlled_property.default_value) / values_range
            inpoint = self._ges_elem.props.in_point
            res = source.set(inpoint, val)
            assert res
            res = source.set(inpoint + self._ges_elem.props.duration, val)
            assert res

    def __create_keyframe_curve(self, bindings=None):
        """Creates required keyframe curve."""
        self.__remove_keyframes()
        if not bindings:
            bindings = [self._ges_elem.get_control_binding(self.__controlled_property.name)]

        if len(bindings) == 1:
            self.keyframe_curve = KeyframeCurve(self.timeline, bindings[0], self._ges_elem)
        else:
            self.keyframe_curve = MultipleKeyframeCurve(self.timeline, bindings, self._ges_elem)

        self.keyframe_curve.connect("enter", self.__curve_enter_cb)
        self.keyframe_curve.connect("leave", self.__curve_leave_cb)
        self.keyframe_curve.set_size_request(self.__width, self.__height)
        self.keyframe_curve.show()
        self.__update_keyframe_curve_visibility()

    def __create_control_binding(self, element):
        """Creates the required ControlBinding and keyframes."""
        if self.__controlled_property:
            element.connect("control-binding-added",
                            self.__control_binding_added_cb)
            binding = \
                element.get_control_binding(self.__controlled_property.name)

            if binding:
                self.__ensure_keyframes(binding)

                return

            source = GstController.InterpolationControlSource()
            source.props.mode = GstController.InterpolationMode.LINEAR
            element.set_control_source(source,
                                       self.__controlled_property.name, "direct")

    def __control_binding_added_cb(self, unused_ges_elem, binding):
        if binding.props.name == self.__controlled_property.name:
            self.__ensure_keyframes(binding)

    def do_draw(self, cr):
        self.propagate_draw(self.__background, cr)

        if self.__previewer:
            self.propagate_draw(self.__previewer, cr)

        if self.keyframe_curve and self.keyframe_curve.is_drawable():
            project = self.timeline.app.project_manager.current_project
            if project.pipeline.get_simple_state() != Gst.State.PLAYING:
                self.propagate_draw(self.keyframe_curve, cr)

    # Callbacks
    def __selected_changed_cb(self, unused_selected, selected):
        if not self.keyframe_curve and self.__controlled_property and \
                selected and len(self.timeline.selection) == 1:
            self.__create_keyframe_curve()

        if self.keyframe_curve:
            self.__update_keyframe_curve_visibility()

        if self.__previewer:
            self.__previewer.set_selected(selected)

    def __update_keyframe_curve_visibility(self):
        """Updates the keyframes widget visibility by adding or removing it."""
        if self._ges_elem.selected and len(self.timeline.selection) == 1:
            if not self.keyframe_curve.get_parent():
                self.add(self.keyframe_curve)
        else:
            self.remove(self.keyframe_curve)

    # Virtual methods
    def _get_previewer(self):
        """Gets a Gtk.Widget to be used as previewer.

        This previewer will be automatically scaled to the width and
        height of the TimelineElement.

        Returns:
            Gtk.Widget: The widget showing thumbnails, waveforms, etc.
        """
        return None

    def _get_background(self):
        """Gets a Gtk.Widget to be used as background.

        Returns:
            Gtk.Widget: The widget identifying the clip type.
        """
        return None

    def _get_default_mixing_property(self):
        """Gets the property controlled by default by the keyframes.

        Returns:
            GObject.ParamSpec: The param spec of the default property.
        """
        return None


class VideoBackground(Gtk.Box):

    def __init__(self):
        Gtk.Box.__init__(self)
        self.get_style_context().add_class("VideoBackground")


class VideoSource(TimelineElement):
    """Widget representing a GES.VideoSource.

    Attributes:
        default_position (dict): The default position (x, y, width, height)
                                 of the VideoSource.
    """

    __gtype_name__ = "PitiviVideoSource"

    def __init__(self, element, timeline):
        super().__init__(element, timeline)

        project = self.timeline.app.project_manager.current_project
        project.connect("video-size-changed",
                        self._project_video_size_changed_cb)

        self.__videoflip = None
        self.__retrieve_project_size()
        self.default_position = self._get_default_position()

        if project.loaded:
            self.__apply_default_position()

        parent = element.get_parent()
        parent.connect("child-added", self.__parent_child_added_cb)
        parent.connect("child-removed", self.__parent_child_removed_cb)

    def __parent_child_added_cb(self, unused_parent, unused_child):
        self.__reset_position()

    def __parent_child_removed_cb(self, unused_parent, child):
        if child == self.__videoflip:
            self.__videoflip = None
            self.__reset_position()
            disconnect_all_by_func(child, self.__track_element_deep_notify_cb)
            disconnect_all_by_func(child, self.__track_element_notify_active_cb)

    def __retrieve_project_size(self):
        project = self.timeline.app.project_manager.current_project

        self._project_width = project.videowidth
        self._project_height = project.videoheight

    def _project_video_size_changed_cb(self, unused_project):
        # GES handles repositionning clips on project size change, make sure to
        # take that into account.
        self.__retrieve_project_size()
        self.default_position = self._get_default_position()

    def __has_default_position(self):
        for name, default_value in self.default_position.items():
            res, value = self._ges_elem.get_child_property(name)
            assert res
            if value != default_value:
                return False

        return True

    def __reset_position(self):
        using_defaults = self.__has_default_position()
        self.__retrieve_project_size()
        self.default_position = self._get_default_position()
        if using_defaults:
            self.debug("Applying default position")
            self.__apply_default_position()
        else:
            self.debug("Not using defaults")

    def __apply_default_position(self):
        video_source = self._ges_elem
        for name, value in self.default_position.items():
            video_source.set_child_property(name, value)

    def _get_default_position(self):
        video_source = self._ges_elem
        sinfo = video_source.get_asset().get_stream_info()

        asset_width = sinfo.get_natural_width()
        asset_height = sinfo.get_natural_height()
        parent = video_source.get_parent()
        if parent and not self.__videoflip:
            for track_element in parent.find_track_elements(
                    None, GES.TrackType.VIDEO, GES.BaseEffect):

                res, unused_videoflip, unused_pspec = track_element.lookup_child(
                    "GstVideoFlip::method")
                if res:
                    self.__videoflip = track_element
                    track_element.connect("deep-notify",
                                          self.__track_element_deep_notify_cb)
                    track_element.connect("notify::active",
                                          self.__track_element_notify_active_cb)

        if self.__videoflip:
            res, method = self.__videoflip.get_child_property("method")
            assert res
            if "clockwise" in method.value_nick and self.__videoflip.props.active:
                asset_width = sinfo.get_natural_height()
                asset_height = sinfo.get_natural_width()

        # Find the biggest size of the video inside the
        # final view (project size) keeping the aspect ratio
        scale = max(self._project_width / asset_width,
                    self._project_height / asset_height)
        if asset_width * scale > self._project_width or \
                asset_height * scale > self._project_height:
            # But make sure it is never bigger than the project!
            scale = min(self._project_width / asset_width,
                        self._project_height / asset_height)

        width = asset_width * scale
        height = asset_height * scale
        x = max(0, (self._project_width - width) / 2)
        y = max(0, (self._project_height - height) / 2)

        self.debug("video scale is %f -> %dx%d", scale, width, height)

        return {"posx": round(x),
                "posy": round(y),
                "width": round(width),
                "height": round(height)}

    def __track_element_deep_notify_cb(self, unused_source, unused_gstelement,
                                       unused_pspec):
        self.__reset_position()

    def __track_element_notify_active_cb(self, unused_track_element,
                                         unused_pspec):
        self.__reset_position()

    def _get_background(self):
        return VideoBackground()


class TitleSource(VideoSource):

    __gtype_name__ = "PitiviTitleSource"

    def _get_default_mixing_property(self):
        for spec in self._ges_elem.list_children_properties():
            if spec.name == "alpha":
                return spec
        return None

    def _get_previewer(self):
        previewer = TitlePreviewer(self._ges_elem)
        previewer.get_style_context().add_class("TitleSource")
        return previewer

    def _get_default_position(self):
        return {"posx": 0,
                "posy": 0,
                "width": self._project_width,
                "height": self._project_height}


class VideoTestSource(VideoSource):

    __gtype_name__ = "PitiviVideoTestSource"

    def _get_default_mixing_property(self):
        for spec in self._ges_elem.list_children_properties():
            if spec.name == "alpha":
                return spec
        return None

    def _get_previewer(self):
        previewer = ImagePreviewer(self._ges_elem, self.timeline.app.settings.previewers_max_cpu)
        return previewer

    def _get_default_position(self):
        return {"posx": 0,
                "posy": 0,
                "width": self._project_width,
                "height": self._project_height}


class VideoUriSource(VideoSource):

    __gtype_name__ = "PitiviUriVideoSource"

    def __init__(self, element, timeline):
        VideoSource.__init__(self, element, timeline)
        self.get_style_context().add_class("VideoUriSource")

    def _get_previewer(self):
        if isinstance(self._ges_elem, GES.ImageSource):
            previewer = ImagePreviewer(self._ges_elem, self.timeline.app.settings.previewers_max_cpu)
        else:
            previewer = VideoPreviewer(self._ges_elem, self.timeline.app.settings.previewers_max_cpu)
        return previewer

    def _get_default_mixing_property(self):
        for spec in self._ges_elem.list_children_properties():
            if spec.name == "alpha":
                return spec
        return None


class AudioBackground(Gtk.Box):

    def __init__(self):
        Gtk.Box.__init__(self)
        self.get_style_context().add_class("AudioBackground")


class AudioUriSource(TimelineElement):

    __gtype_name__ = "PitiviAudioUriSource"

    def __init__(self, element, timeline):
        TimelineElement.__init__(self, element, timeline)
        self.get_style_context().add_class("AudioUriSource")

    def _get_previewer(self):
        previewer = AudioPreviewer(self._ges_elem, self.timeline.app.settings.previewers_max_cpu)
        return previewer

    def _get_background(self):
        return AudioBackground()

    def _get_default_mixing_property(self):
        for spec in self._ges_elem.list_children_properties():
            if spec.name == "volume":
                return spec
        return None


class TrimHandle(Gtk.EventBox, Loggable):

    __gtype_name__ = "PitiviTrimHandle"

    SELECTED_WIDTH = 5
    DEFAULT_WIDTH = 1
    PIXBUF = None

    def __init__(self, clip, edge):
        Gtk.EventBox.__init__(self)
        Loggable.__init__(self)

        self.clip = clip
        self.edge = edge

        self.get_style_context().add_class("Trimbar")
        if edge == GES.Edge.EDGE_END:
            css_class = "right"
        else:
            css_class = "left"
        self.get_style_context().add_class(css_class)

        self.props.valign = Gtk.Align.FILL
        self.shrink()
        if edge == GES.Edge.EDGE_END:
            self.props.halign = Gtk.Align.END
        else:
            self.props.halign = Gtk.Align.START

    def do_draw(self, cr):
        Gtk.EventBox.do_draw(self, cr)
        if TrimHandle.PIXBUF is None:
            TrimHandle.PIXBUF = GdkPixbuf.Pixbuf.new_from_file(
                os.path.join(get_pixmap_dir(), "trimbar-focused.png"))
        Gdk.cairo_set_source_pixbuf(cr, TrimHandle.PIXBUF, 10, 10)

    def enlarge(self):
        self.props.width_request = TrimHandle.SELECTED_WIDTH
        if self.props.window:
            self.props.window.set_cursor(CURSORS[self.edge])

    def shrink(self):
        self.props.width_request = TrimHandle.DEFAULT_WIDTH
        if self.props.window:
            self.props.window.set_cursor(NORMAL_CURSOR)


class Clip(Gtk.EventBox, Zoomable, Loggable):

    __gtype_name__ = "PitiviClip"

    def __init__(self, layer, ges_clip):
        Gtk.EventBox.__init__(self)
        Zoomable.__init__(self)
        Loggable.__init__(self)

        name = ges_clip.get_name()
        self.set_name(name)
        self.get_accessible().set_name(name)

        self._elements_container = None
        self.left_handle = None
        self.right_handle = None
        self.handles = []
        self.z_order = -1
        self.timeline = layer.timeline
        self.app = layer.app

        self.ges_clip = ges_clip
        self.ges_clip.ui = self
        self.ges_clip.selected = Selected()

        self.audio_widget = None
        self.video_widget = None

        self._setup_widget()
        self.__force_position_update = True

        for ges_timeline_element in self.ges_clip.get_children(False):
            self._add_child(ges_timeline_element)
            self.__connect_to_child(ges_timeline_element)

        # Connect to Widget signals.
        self.connect("button-release-event", self._button_release_event_cb)
        self.connect("event", self._event_cb)

        # Connect to GES signals.
        self.ges_clip.connect("notify::start", self._start_changed_cb)
        self.ges_clip.connect("notify::inpoint", self._start_changed_cb)
        self.ges_clip.connect("notify::duration", self._duration_changed_cb)
        self.ges_clip.connect("notify::layer", self._layer_changed_cb)

        self.ges_clip.connect_after("child-added", self._child_added_cb)
        self.ges_clip.connect_after("child-removed", self._child_removed_cb)

        # To be able to receive effects dragged on clips.
        self.drag_dest_set(0, [EFFECT_TARGET_ENTRY], Gdk.DragAction.COPY)
        self.connect("drag-drop", self.__drag_drop_cb)

    @property
    def layer(self):
        ges_layer = self.ges_clip.props.layer
        return ges_layer.ui if ges_layer else None

    def __drag_drop_cb(self, widget, context, x, y, timestamp):
        success = False

        target = self.drag_dest_find_target(context, None)
        if not target:
            return False

        if target.name() == EFFECT_TARGET_ENTRY.target:
            self.info("Adding effect %s", self.timeline.drop_data)
            self.timeline.selection.set_selection([self.ges_clip], SELECT)
            self.app.gui.editor.switch_context_tab(self.ges_clip)

            effect_info = self.app.effects.get_info(self.timeline.drop_data)
            pipeline = self.timeline.ges_timeline.get_parent()
            with self.app.action_log.started("add effect",
                                             finalizing_action=CommitTimelineFinalizingAction(pipeline),
                                             toplevel=True):
                self.add_effect(effect_info)
            self.timeline.clean_drop_data()
            success = True

        Gtk.drag_finish(context, success, False, timestamp)

        return success

    def add_effect(self, effect_info):
        """Adds the specified effect if it can be applied to the clip.

        Args:
            effect_info (EffectInfo): The effect to add.
        """
        factory_name = effect_info.effect_name
        if factory_name in ALLOWED_ONLY_ONCE_EFFECTS:
            for effect in self.ges_clip.find_track_elements(None, GES.TrackType.VIDEO,
                                                            GES.BaseEffect):
                for elem in effect.get_nleobject().iterate_recurse():
                    if elem.get_factory().get_name() == factory_name:
                        self.error("Not adding %s as it would be duplicate"
                                   " and this is not allowed.", factory_name)
                        # TODO Let the user know about why it did not work.
                        return effect

        for track_element in self.ges_clip.get_children(False):
            if effect_info.good_for_track_element(track_element):
                # Actually add the effect
                effect = GES.Effect.new(effect_info.bin_description)
                self.ges_clip.add(effect)
                return effect
        return None

    def update_position(self):
        layer = self.layer
        if not layer or layer != self.get_parent():
            # Things are not settled yet.
            return

        start = self.ges_clip.props.start
        duration = self.ges_clip.props.duration
        x = self.ns_to_pixel(start)
        # The calculation of the width assumes that the start is always
        # int(pixels_float). In that case, the rounding can add up and a pixel
        # might be lost if we ignore the start of the clip.
        width = self.ns_to_pixel(start + duration) - x

        parent_height = layer.props.height_request
        y = 0
        height = parent_height
        has_video = self.ges_clip.find_track_elements(None, GES.TrackType.VIDEO, GObject.TYPE_NONE)
        has_audio = self.ges_clip.find_track_elements(None, GES.TrackType.AUDIO, GObject.TYPE_NONE)
        if not has_video or not has_audio:
            if layer.media_types == (GES.TrackType.AUDIO | GES.TrackType.VIDEO):
                height = parent_height / 2
                if not has_video:
                    y = height

        if self.__force_position_update or \
                x != self._current_x or \
                y != self._current_y or \
                width != self._current_width or \
                parent_height != self._current_parent_height or \
                layer != self._current_parent:

            layer.move(self, x, y)
            self.set_size_request(width, height)

            elements = self._elements_container.get_children()
            for child in elements:
                child.set_size(width, height / len(elements))

            self.__force_position_update = False
            # pylint: disable=attribute-defined-outside-init
            self._current_x = x
            self._current_y = y
            self._current_width = width
            self._current_parent_height = parent_height
            self._current_parent = layer

    def _setup_widget(self):
        pass

    def _add_trim_handles(self):
        overlay = Gtk.Overlay()
        self.add(overlay)

        self._elements_container = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        overlay.add_overlay(self._elements_container)

        self.left_handle = TrimHandle(self, GES.Edge.EDGE_START)
        overlay.add_overlay(self.left_handle)

        self.right_handle = TrimHandle(self, GES.Edge.EDGE_END)
        overlay.add_overlay(self.right_handle)

        self.handles.append(self.left_handle)
        self.handles.append(self.right_handle)

    def shrink_trim_handles(self):
        for handle in self.handles:
            handle.shrink()

    def do_map(self):
        Gtk.EventBox.do_map(self)
        self.update_position()

    def _button_release_event_cb(self, unused_widget, event):
        if self.timeline.got_dragged:
            # This means a drag & drop operation just finished and
            # this button-release-event should be ignored.
            self.timeline.got_dragged = False
            return False

        res, button = event.get_button()
        if res and not button == 1:
            # Only the left mouse button selects.
            return False

        mode = SELECT
        if self.timeline.get_parent().control_mask:
            if not self.get_state_flags() & Gtk.StateFlags.SELECTED:
                mode = SELECT_ADD
            else:
                mode = UNSELECT
            clicked_layer, click_pos = self.timeline.get_clicked_layer_and_pos(event)
            self.timeline.set_selection_meta_info(clicked_layer, click_pos, mode)
        else:
            self.app.gui.editor.switch_context_tab(self.ges_clip)

        parent = self.ges_clip.get_toplevel_parent()
        if parent is self.ges_clip:
            selection = [self.ges_clip]
        else:
            selection = [elem for elem in parent.get_children(True)
                         if isinstance(elem, (GES.SourceClip, GES.TransitionClip))]
        self.timeline.selection.set_selection(selection, mode)

        return False

    def release(self):
        for child in self.ges_clip.get_children(True):
            self.__disconnect_from_child(child)

        disconnect_all_by_func(self.ges_clip, self._start_changed_cb)
        disconnect_all_by_func(self.ges_clip, self._duration_changed_cb)
        disconnect_all_by_func(self.ges_clip, self._layer_changed_cb)
        disconnect_all_by_func(self.ges_clip, self._child_added_cb)
        disconnect_all_by_func(self.ges_clip, self._child_removed_cb)

    def __show_handles(self):
        for handle in self.handles:
            handle.show()

    def __hide_handles(self):
        for handle in self.handles:
            handle.hide()

    def _event_cb(self, element, event):
        if (event.type == Gdk.EventType.ENTER_NOTIFY and
                event.mode == Gdk.CrossingMode.NORMAL and
                not self.timeline.scrubbing):
            set_children_state_recurse(self, Gtk.StateFlags.PRELIGHT)
            for handle in self.handles:
                handle.enlarge()
        elif (event.type == Gdk.EventType.LEAVE_NOTIFY and
              event.mode == Gdk.CrossingMode.NORMAL):
            unset_children_state_recurse(self, Gtk.StateFlags.PRELIGHT)
            for handle in self.handles:
                handle.shrink()

        return False

    def _start_changed_cb(self, clip, pspec):
        self.update_position()

    def _duration_changed_cb(self, clip, pspec):
        self.update_position()

    def _layer_changed_cb(self, ges_clip, pspec):
        self.update_position()

    def __disconnect_from_child(self, child):
        if child.ui:
            child.ui.release()

    def __connect_to_child(self, child):
        if child.ui:
            child.ui.connect("curve-enter", self.__curve_enter_cb)
            child.ui.connect("curve-leave", self.__curve_leave_cb)

    def __curve_enter_cb(self, unused_keyframe_curve):
        self.__hide_handles()

    def __curve_leave_cb(self, unused_keyframe_curve):
        self.__show_handles()

    def _add_child(self, ges_timeline_element):
        ges_timeline_element.selected = Selected()
        ges_timeline_element.ui = None

    def _child_added_cb(self, unused_ges_clip, ges_timeline_element):
        self.__force_position_update = True
        self._add_child(ges_timeline_element)
        self.__connect_to_child(ges_timeline_element)
        self.update_position()

    def _remove_child(self, ges_timeline_element):
        pass

    def _child_removed_cb(self, unused_ges_clip, ges_timeline_element):
        self.__force_position_update = True
        self.__disconnect_from_child(ges_timeline_element)
        self._remove_child(ges_timeline_element)
        self.update_position()


class SourceClip(Clip):
    __gtype_name__ = "PitiviSourceClip"

    def __init__(self, layer, ges_clip):
        Clip.__init__(self, layer, ges_clip)

    def _setup_widget(self):
        self._add_trim_handles()

        self.get_style_context().add_class("Clip")

    def _remove_child(self, ges_timeline_element):
        if ges_timeline_element.ui:
            self._elements_container.remove(ges_timeline_element.ui)
            ges_timeline_element.ui = None


class UriClip(SourceClip):
    __gtype_name__ = "PitiviUriClip"

    def __init__(self, layer, ges_clip):
        SourceClip.__init__(self, layer, ges_clip)
        self.get_style_context().add_class("UriClip")
        self.props.has_tooltip = True

    def do_query_tooltip(self, x, y, keyboard_mode, tooltip):
        tooltip.set_markup(filename_from_uri(
            self.ges_clip.get_asset().props.id))

        return True

    def _add_child(self, ges_timeline_element):
        SourceClip._add_child(self, ges_timeline_element)

        if not isinstance(ges_timeline_element, GES.Source):
            return

        if ges_timeline_element.get_track_type() == GES.TrackType.AUDIO:
            self.audio_widget = AudioUriSource(ges_timeline_element, self.timeline)
            ges_timeline_element.ui = self.audio_widget
            self._elements_container.pack_end(self.audio_widget, True, False, 0)
            self.audio_widget.set_visible(True)
        elif ges_timeline_element.get_track_type() == GES.TrackType.VIDEO:
            self.video_widget = VideoUriSource(ges_timeline_element, self.timeline)
            ges_timeline_element.ui = self.video_widget
            self._elements_container.pack_start(self.video_widget, True, False, 0)
            self.video_widget.set_visible(True)


class TestClip(SourceClip):
    __gtype_name__ = "PitiviTestClip"

    def __init__(self, layer, ges_clip):
        SourceClip.__init__(self, layer, ges_clip)
        self.get_style_context().add_class("TestClip")

    def _add_child(self, ges_timeline_element):
        SourceClip._add_child(self, ges_timeline_element)

        if not isinstance(ges_timeline_element, GES.Source):
            return

        if ges_timeline_element.get_track_type() == GES.TrackType.VIDEO:
            self.video_widget = VideoTestSource(ges_timeline_element, self.timeline)
            ges_timeline_element.ui = self.video_widget
            self._elements_container.pack_start(self.video_widget, True, False, 0)
            self.video_widget.set_visible(True)


class TitleClip(SourceClip):
    __gtype_name__ = "PitiviTitleClip"

    def __init__(self, layer, ges_clip):
        SourceClip.__init__(self, layer, ges_clip)
        self.get_style_context().add_class("TitleClip")

    def _add_child(self, ges_timeline_element):
        SourceClip._add_child(self, ges_timeline_element)

        if not isinstance(ges_timeline_element, GES.Source):
            return

        if ges_timeline_element.get_track_type() == GES.TrackType.VIDEO:
            self.video_widget = TitleSource(ges_timeline_element, self.timeline)
            ges_timeline_element.ui = self.video_widget
            self._elements_container.pack_start(self.video_widget, True, False, 0)
            self.video_widget.set_visible(True)


class TransitionClip(Clip):

    __gtype_name__ = "PitiviTransitionClip"

    def __init__(self, layer, ges_clip):
        self.__has_video = False

        Clip.__init__(self, layer, ges_clip)

        if self.__has_video:
            self.z_order = 1
        else:
            self.z_order = 0

        self.get_style_context().add_class("TransitionClip")

        # In the case of TransitionClips, we are the only container
        self._add_trim_handles()

        self.props.has_tooltip = True

    def do_query_tooltip(self, x, y, keyboard_mode, tooltip):
        if self.__has_video:
            markup = str(self.ges_clip.props.vtype.value_nick)
        else:
            markup = _("Audio crossfade")
        tooltip.set_text(markup)

        return True

    def _add_child(self, ges_timeline_element):
        Clip._add_child(self, ges_timeline_element)

        if not isinstance(ges_timeline_element, GES.VideoTransition):
            return

        self.z_order = 1
        self.set_sensitive(True)
        self.__has_video = True
        ges_timeline_element.selected.connect("selected-changed", self._selected_changed_cb, ges_timeline_element)

    def _selected_changed_cb(self, unused_selected, selected, ges_timeline_element):
        if selected:
            self.app.gui.editor.trans_list.activate(ges_timeline_element)
        else:
            self.app.gui.editor.trans_list.deactivate()


GES_TYPE_UI_TYPE = {
    GES.UriClip.__gtype__: UriClip,
    GES.TitleClip.__gtype__: TitleClip,
    GES.TransitionClip.__gtype__: TransitionClip,
    GES.TestClip.__gtype__: TestClip
}



# -*- coding: utf-8 -*-
# Pitivi Developer Console
# Copyright (c) 2017-2018, Fabian Orccon <cfoch.fabian@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Python console for inspecting and interacting with Pitivi and the project."""
import sys
from gettext import gettext as _

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import Peas

from pitivi.dialogs.prefs import PreferencesDialog
from plugins.console.utils import Namespace
from plugins.console.widgets import ConsoleWidget


class PitiviNamespace(Namespace):
    """Easy to shape Python namespace."""

    def __init__(self, app):
        Namespace.__init__(self)
        self._app = app

    @property
    @Namespace.shortcut
    def app(self):
        """The Pitivi instance."""
        return self._app

    @property
    @Namespace.shortcut
    def plugin_manager(self):
        """The Plugin Manager instance."""
        return self._app.plugin_manager

    @property
    @Namespace.shortcut
    def project(self):
        """The current project."""
        return self._app.project_manager.current_project

    @property
    @Namespace.shortcut
    def timeline(self):
        """The GES.Timeline of the current project."""
        return self._app.gui.editor.timeline_ui.timeline.ges_timeline


class Console(GObject.GObject, Peas.Activatable):
    """Plugin which adds a Python console for development purposes."""

    __gtype_name__ = "ConsolePlugin"
    object = GObject.Property(type=GObject.Object)

    DEFAULT_COLOR = Gdk.RGBA(1.0, 1.0, 1.0, 1.0)
    DEFAULT_STDERR_COLOR = Gdk.RGBA(0.96, 0.47, 0.0, 1.0)
    DEFAULT_STDOUT_COLOR = Gdk.RGBA(1.0, 1.0, 1.0, 1.0)
    DEFAULT_FONT = Pango.FontDescription.from_string("Monospace Regular 12")

    def __init__(self):
        GObject.GObject.__init__(self)
        self.window = None
        self.terminal = None
        self.menu_item = None
        self.app = None

        # Set prompt.
        sys.ps1 = ">>> "
        sys.ps2 = "... "

    def do_activate(self):
        api = self.object
        self.app = api.app
        self.app.settings.add_config_section("console")

        self.app.settings.add_config_option(attrname="consoleColor",
                                            section="console",
                                            key="console-color",
                                            notify=True,
                                            default=Console.DEFAULT_COLOR)

        self.app.settings.add_config_option(attrname="consoleErrorColor",
                                            section="console",
                                            key="console-error-color",
                                            notify=True,
                                            default=Console.DEFAULT_STDERR_COLOR)

        self.app.settings.add_config_option(attrname="consoleOutputColor",
                                            section="console",
                                            key="console-output-color",
                                            notify=True,
                                            default=Console.DEFAULT_STDOUT_COLOR)

        self.app.settings.add_config_option(attrname="consoleFont",
                                            section="console",
                                            key="console-font",
                                            notify=True,
                                            default=Console.DEFAULT_FONT.to_string())

        self.app.settings.reload_attribute_from_file("console", "consoleColor")
        self.app.settings.reload_attribute_from_file("console",
                                                     "consoleErrorColor")
        self.app.settings.reload_attribute_from_file("console",
                                                     "consoleOutputColor")
        self.app.settings.reload_attribute_from_file("console", "consoleFont")

        PreferencesDialog.add_section("console", _("Console"))
        PreferencesDialog.add_color_preference(attrname="consoleColor",
                                               label=_("Color"),
                                               description=None,
                                               section="console")
        PreferencesDialog.add_color_preference(attrname="consoleErrorColor",
                                               # Translators: The color of the content from stderr.
                                               label=_("Standard error color"),
                                               description=None,
                                               section="console")
        PreferencesDialog.add_color_preference(attrname="consoleOutputColor",
                                               # Translators: The color of the content from stdout.
                                               label=_("Standard output color"),
                                               description=None,
                                               section="console")
        PreferencesDialog.add_font_preference(attrname="consoleFont",
                                              label=_("Font"),
                                              description=None,
                                              section="console")

        open_action = Gio.SimpleAction.new("open_console", None)
        open_action.connect("activate", self.__menu_item_activate_cb)
        self.app.add_action(open_action)
        self.app.shortcuts.add("app.open_console", ["<Primary>d"], open_action, _("Developer Console"))

        self._setup_dialog()
        self.add_menu_item()
        self.menu_item.show()

    def do_deactivate(self):
        PreferencesDialog.remove_section("console")
        self.window.destroy()
        self.remove_menu_item()
        self.window = None
        self.terminal = None
        self.app = None

    def add_menu_item(self):
        """Inserts a menu item into the Pitivi menu."""
        menu = self.app.gui.editor.builder.get_object("menu_box")
        self.menu_item = Gtk.ModelButton.new()
        self.menu_item.props.text = _("Developer Console")
        self.menu_item.set_action_name("app.open_console")

        menu.add(self.menu_item)

    def remove_menu_item(self):
        """Removes a menu item from the Pitivi menu."""
        self.app.remove_action("open_console")
        menu = self.app.gui.editor.builder.get_object("menu_box")
        menu.remove(self.menu_item)
        self.menu_item = None

    def _setup_dialog(self):
        namespace = PitiviNamespace(self.app)
        self.window = Gtk.Window()
        welcome_message = "".join(self._create_welcome_message(namespace))
        self.terminal = ConsoleWidget(namespace, welcome_message)
        self.terminal.connect("eof", self.__eof_cb)

        self._init_colors()
        self.terminal.set_font(self.app.settings.consoleFont)
        self._connect_settings_signals()

        self.window.set_default_size(600, 400)
        self.window.set_title(_("Pitivi Console"))
        self.window.connect("delete-event", self.__delete_event_cb)
        self.window.add(self.terminal)

    def _create_welcome_message(self, namespace):
        console_plugin_info = self.app.plugin_manager.get_plugin_info("console")
        name = console_plugin_info.get_name()
        version = console_plugin_info.get_version() or ""
        yield "%s %s\n\n" % (name, version)
        yield console_plugin_info.get_help_uri()
        yield "\n\n"
        yield _("You can use the following shortcuts:")
        yield "\n"
        for shortcut in namespace.get_shortcuts():
            yield " - %s\n" % shortcut
        yield "\n"
        yield _("Type \"{help}(<command>)\" for more information.").format(help="help")
        yield "\n\n"

    def _init_colors(self):
        """Sets the colors from Pitivi settings."""
        self.terminal.set_stderr_color(self.app.settings.consoleErrorColor)
        self.terminal.set_stdout_color(self.app.settings.consoleOutputColor)
        self.terminal.set_color(self.app.settings.consoleColor)

    def _connect_settings_signals(self):
        """Connects the settings' signals."""
        self.app.settings.connect("consoleColorChanged", self.__color_changed_cb)
        self.app.settings.connect("consoleErrorColorChanged",
                                  self.__error_color_changed_cb)
        self.app.settings.connect("consoleOutputColorChanged",
                                  self.__output_color_changed_cb)
        self.app.settings.connect("consoleFontChanged", self.__font_changed_cb)

    def __color_changed_cb(self, settings):
        if self.terminal:
            self.terminal.set_color(settings.consoleColor)

    def __error_color_changed_cb(self, settings):
        if self.terminal:
            self.terminal.set_stderr_color(settings.consoleErrorColor)

    def __output_color_changed_cb(self, settings):
        if self.terminal:
            self.terminal.set_stdout_color(settings.consoleOutputColor)

    def __font_changed_cb(self, settings):
        if self.terminal:
            self.terminal.set_font(settings.consoleFont)

    def __menu_item_activate_cb(self, unused_data, unused_param):
        self.window.show_all()
        self.window.set_keep_above(True)

    def __eof_cb(self, unused_widget):
        self.window.hide()
        return True

    def __delete_event_cb(self, unused_widget, unused_data):
        return self.window.hide_on_delete()



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the pitivi.interactiveintro module."""
# pylint: disable=protected-access,unused-argument
from gi.repository import GLib

from tests import common


class TestInteractiveIntro(common.TestCase):

    def test_overview(self):
        app = common.create_pitivi()
        app.gui = None
        app.create_main_window()
        intro = app.gui.editor.intro
        app.project_manager.new_blank_project()

        intro.intro_action.activate()

        def timeout_add(timeout, func, *args):
            func(*args)

        original_timeout_add = GLib.timeout_add
        GLib.timeout_add = timeout_add
        try:
            intro._overview_button_clicked_cb(None)
        finally:
            GLib.timeout_add = original_timeout_add

        self.assertListEqual(intro.tips, [])



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
# Copyright (c) 2009, Alessandro Decina <alessandro.d@gmail.com>
# Copyright (c) 2012, Jean-François Fortin Tam <nekohayo@gmail.com>
# Copyright (c) 2020, Abhishek Kumar Singh <kumarsinghabhishek59@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import os
import subprocess
import sys
import time
from enum import IntEnum
from gettext import gettext as _
from gettext import ngettext
from hashlib import md5

import cairo
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GES
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstPbutils
from gi.repository import Gtk
from gi.repository import Pango

from pitivi.configure import get_pixmap_dir
from pitivi.configure import get_ui_dir
from pitivi.dialogs.clipmediaprops import ClipMediaPropsDialog
from pitivi.dialogs.filelisterrordialog import FileListErrorDialog
from pitivi.mediafilespreviewer import PreviewWidget
from pitivi.settings import GlobalSettings
from pitivi.timeline.previewers import AssetPreviewer
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import disconnect_all_by_func
from pitivi.utils.misc import path_from_uri
from pitivi.utils.misc import PathWalker
from pitivi.utils.misc import quote_uri
from pitivi.utils.misc import show_user_manual
from pitivi.utils.proxy import get_proxy_target
from pitivi.utils.proxy import ProxyingStrategy
from pitivi.utils.ui import beautify_asset
from pitivi.utils.ui import beautify_eta
from pitivi.utils.ui import FILE_TARGET_ENTRY
from pitivi.utils.ui import filter_unsupported_media_files
from pitivi.utils.ui import fix_infobar
from pitivi.utils.ui import info_name
from pitivi.utils.ui import LARGE_THUMB_WIDTH
from pitivi.utils.ui import PADDING
from pitivi.utils.ui import SMALL_THUMB_WIDTH
from pitivi.utils.ui import SPACING
from pitivi.utils.ui import URI_TARGET_ENTRY


class ViewType(IntEnum):
    """How the assets can be displayed."""

    LIST = 1
    ICON = 2


GlobalSettings.add_config_section('clip-library')
GlobalSettings.add_config_option('lastImportFolder',
                                 section='clip-library',
                                 key='last-folder',
                                 environment='PITIVI_IMPORT_FOLDER',
                                 default=os.path.expanduser("~"))
GlobalSettings.add_config_option('closeImportDialog',
                                 section='clip-library',
                                 key='close-import-dialog-after-import',
                                 default=True)
GlobalSettings.add_config_option('last_clip_view',
                                 section='clip-library',
                                 key='last-clip-view',
                                 type_=str,
                                 default=ViewType.ICON.name)


class AssetStoreItem(GObject.GObject):
    """Data for displaying an asset."""

    def __init__(self, asset, thumb_decorator):
        GObject.GObject.__init__(self)
        self.icon_64 = thumb_decorator.small_thumb
        self.icon_128 = thumb_decorator.large_thumb
        self.infotext = beautify_asset(asset)
        self.asset = asset
        self.uri = self.asset.props.id
        self.search_text = info_name(asset)
        self.thumb_decorator = thumb_decorator

        # Fetch or Register tags
        tags = asset.get_meta("pitivi::tags")
        if not tags:
            self.tags = set()
            self.asset.register_meta(GES.MetaFlag.READWRITE, "pitivi::tags", "")
        else:
            self.tags = set(tags.split(","))


class TagState(IntEnum):
    """How the tag is associated with assets under selection."""

    PRESENT = 1
    INCONSISTENT = 2
    ABSENT = 3


class TagStoreItem(GObject.GObject):
    """Data for displaying a Tag in the list."""

    def __init__(self, name: str, initial_state: TagState):
        GObject.GObject.__init__(self)
        self.name = name
        self.initial_state = initial_state


class OptimizeOption(IntEnum):
    UNSUPPORTED_ASSETS = 0
    ALL = 1


class FileChooserExtraWidget(Gtk.Box, Loggable):

    def __init__(self, app):
        Loggable.__init__(self)
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
        self.app = app

        self.__keep_open_check = Gtk.CheckButton(label=_("Keep dialog open"))
        self.__keep_open_check.props.valign = Gtk.Align.START
        self.__keep_open_check.set_tooltip_text(_("When importing files keep the dialog open"))
        self.__keep_open_check.set_active(not self.app.settings.closeImportDialog)
        self.pack_start(self.__keep_open_check, expand=False, fill=False, padding=0)

        self.hq_proxy_check = Gtk.CheckButton.new()
        # Translators: Create optimized media for unsupported files.
        self.hq_proxy_check.set_label(_("Optimize:"))
        self.hq_proxy_check.connect("toggled", self._hq_proxy_check_cb)

        self.hq_combo = Gtk.ComboBoxText.new()
        self.hq_combo.insert_text(OptimizeOption.UNSUPPORTED_ASSETS, _("Unsupported assets"))
        self.hq_combo.insert_text(OptimizeOption.ALL, _("All"))
        self.hq_combo.props.active = OptimizeOption.UNSUPPORTED_ASSETS
        self.hq_combo.set_sensitive(False)

        self.help_button = Gtk.Button()
        self.__update_help_button()
        self.help_button.props.relief = Gtk.ReliefStyle.NONE
        self.help_button.connect("clicked", self._help_button_clicked_cb)

        self.scaled_proxy_check = Gtk.CheckButton.new()
        self.__update_scaled_proxy_check()
        self.scaled_proxy_check.connect("toggled", self._scaled_proxy_check_cb)

        self.project_settings_label = Gtk.Label()
        self.project_settings_label.set_markup("<a href='#'>%s</a>" % _("Project Settings"))
        self.project_settings_label.connect("activate-link", self._target_res_cb)

        proxy_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        hq_proxy_row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        hq_proxy_row.pack_start(self.hq_proxy_check, expand=False, fill=False, padding=0)
        hq_proxy_row.pack_start(self.hq_combo, expand=False, fill=False, padding=PADDING)
        hq_proxy_row.pack_start(self.help_button, expand=False, fill=False, padding=SPACING)
        proxy_box.pack_start(hq_proxy_row, expand=False, fill=False, padding=0)

        row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        row.pack_start(self.scaled_proxy_check, expand=False, fill=False, padding=0)
        row.pack_start(self.project_settings_label, expand=False, fill=False, padding=SPACING)
        proxy_box.pack_start(row, expand=False, fill=False, padding=0)

        self.pack_start(proxy_box, expand=False, fill=False, padding=SPACING * 2)

        self.show_all()

        size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.VERTICAL)
        size_group.add_widget(self.__keep_open_check)
        size_group.add_widget(hq_proxy_row)

        if self.app.settings.proxying_strategy == ProxyingStrategy.AUTOMATIC:
            self.hq_proxy_check.set_active(True)
            self.hq_combo.set_sensitive(True)
            self.hq_combo.props.active = OptimizeOption.UNSUPPORTED_ASSETS
        elif self.app.settings.proxying_strategy == ProxyingStrategy.ALL:
            self.hq_proxy_check.set_active(True)
            self.hq_combo.set_sensitive(True)
            self.hq_combo.props.active = OptimizeOption.ALL

        if self.app.settings.auto_scaling_enabled:
            self.scaled_proxy_check.set_active(True)

    def _hq_proxy_check_cb(self, check_button):
        active = check_button.get_active()
        self.hq_combo.set_sensitive(active)
        self.__update_help_button()

        if not active:
            self.scaled_proxy_check.set_active(False)

    def __update_help_button(self):
        if self.hq_proxy_check.get_active():
            icon = "question-round-symbolic"
        else:
            icon = "warning-symbolic"
        image = Gtk.Image.new_from_icon_name(icon, Gtk.IconSize.BUTTON)
        self.help_button.set_image(image)

    def _help_button_clicked_cb(self, unused_button):
        show_user_manual("importing")

    def _scaled_proxy_check_cb(self, unused_button):
        if self.scaled_proxy_check.get_active():
            self.hq_combo.props.active = OptimizeOption.UNSUPPORTED_ASSETS
            self.hq_proxy_check.set_active(True)

    def _target_res_cb(self, label_widget, unused_uri):
        self.app.gui.editor.show_project_settings_dialog()
        self.__update_scaled_proxy_check()

    def __update_scaled_proxy_check(self):
        target_width = self.app.project_manager.current_project.scaled_proxy_width
        target_height = self.app.project_manager.current_project.scaled_proxy_height
        self.scaled_proxy_check.set_label(_("Scale assets larger than %s×%s px.") % (target_width, target_height))

    def save_values(self):
        self.app.settings.closeImportDialog = not self.__keep_open_check.get_active()

        if self.hq_proxy_check.get_active():
            if self.hq_combo.props.active == OptimizeOption.UNSUPPORTED_ASSETS:
                self.app.settings.proxying_strategy = ProxyingStrategy.AUTOMATIC
            else:
                self.app.settings.proxying_strategy = ProxyingStrategy.ALL
        else:
            assert not self.scaled_proxy_check.get_active()
            self.app.settings.proxying_strategy = ProxyingStrategy.NOTHING

        self.app.settings.auto_scaling_enabled = self.scaled_proxy_check.get_active()


class AssetThumbnail(GObject.Object, Loggable):
    """Provider of decorated thumbnails for an asset.

    The small_thumb and large_thumb fields hold the thumbs decorated
    according to the status of the asset.
    """

    __gsignals__ = {
        # Emitted when small_thumb and large_thumb changed.
        "thumb-updated": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    EMBLEMS = {}
    PROXIED = "asset-proxied"
    SCALED = "asset-scaled"
    NO_PROXY = "no-proxy"
    IN_PROGRESS = "asset-proxy-in-progress"
    ASSET_PROXYING_ERROR = "asset-proxying-error"
    UNSUPPORTED = "asset-unsupported"

    DEFAULT_ALPHA = 255

    icons_by_name = {}

    for status in [PROXIED, SCALED, IN_PROGRESS, ASSET_PROXYING_ERROR, UNSUPPORTED]:
        EMBLEMS[status] = GdkPixbuf.Pixbuf.new_from_file_at_size(
            os.path.join(get_pixmap_dir(), "%s.svg" % status), 64, 64)

    def __init__(self, asset, proxy_manager):
        GObject.Object.__init__(self)
        Loggable.__init__(self)
        self.__asset = asset
        self.proxy_manager = proxy_manager
        self.__previewer = None
        self.small_thumb = None
        self.large_thumb = None
        self.refresh()

    def refresh(self):
        """Updates the shown icon. To be called when a new icon is available."""
        self.src_small, self.src_large = self.__get_thumbnails()
        self.decorate()

    def disregard_previewer(self):
        if self.__previewer:
            self.__previewer.disconnect_by_func(self.__done_cb)
            self.__previewer.stop_generation()
            self.__previewer = None

        self.refresh()
        self.emit("thumb-updated")

    def __get_thumbnails(self):
        """Gets the base source thumbnails.

        Returns:
            List[GdkPixbuf.Pixbuf]: The small thumbnail and the large thumbnail
            to be decorated.
        """
        video_streams = [
            stream_info
            for stream_info in self.__asset.get_info().get_stream_list()
            if isinstance(stream_info, GstPbutils.DiscovererVideoInfo)]
        if video_streams:
            # Check if the files have thumbnails in the user's cache directory.
            real_uri = get_proxy_target(self.__asset).props.id
            small_thumb, large_thumb = self.get_thumbnails_from_xdg_cache(real_uri)
            if not small_thumb:
                if self.__asset.is_image():
                    path = Gst.uri_get_location(real_uri)
                    try:
                        pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
                        width = pixbuf.props.width
                        height = pixbuf.props.height
                        small_thumb = pixbuf.scale_simple(
                            SMALL_THUMB_WIDTH,
                            SMALL_THUMB_WIDTH * height / width,
                            GdkPixbuf.InterpType.BILINEAR)
                        large_thumb = pixbuf.scale_simple(
                            LARGE_THUMB_WIDTH,
                            LARGE_THUMB_WIDTH * height / width,
                            GdkPixbuf.InterpType.BILINEAR)
                    except GLib.Error as error:
                        self.debug("Failed loading thumbnail because: %s", error)
                        small_thumb, large_thumb = self.__get_icons("image-x-generic")
                else:
                    # Build or reuse a ThumbnailCache.
                    if not self.__previewer:
                        self.__previewer = AssetPreviewer(self.__asset, 90)
                        self.__previewer.connect("done", self.__done_cb)
                    small_thumb = self.__previewer.thumb_cache.get_preview_thumbnail()
                    if not small_thumb:
                        # We'll be notified when the thumbnail is available.
                        small_thumb, large_thumb = self.__get_icons("video-x-generic")
                    else:
                        width = small_thumb.props.width
                        height = small_thumb.props.height
                        large_thumb = small_thumb.scale_simple(
                            LARGE_THUMB_WIDTH,
                            LARGE_THUMB_WIDTH * height / width,
                            GdkPixbuf.InterpType.BILINEAR)
                        if width > SMALL_THUMB_WIDTH:
                            small_thumb = small_thumb.scale_simple(
                                SMALL_THUMB_WIDTH,
                                SMALL_THUMB_WIDTH * height / width,
                                GdkPixbuf.InterpType.BILINEAR)
        else:
            small_thumb, large_thumb = self.__get_icons("audio-x-generic")
        return small_thumb, large_thumb

    def __done_cb(self, unused_asset_previewer):
        """Handles the done signal of our AssetPreviewer."""
        self.refresh()
        self.emit("thumb-updated")

    @staticmethod
    def get_asset_thumbnails_path(real_uri):
        """Gets normal & large thumbnail path for the asset in the XDG cache.

        Returns:
            List[str]: The path of normal thumbnail and large thumbnail.
        """
        quoted_uri = quote_uri(real_uri)
        thumbnail_hash = md5(quoted_uri.encode()).hexdigest()
        thumb_dir = os.path.join(GLib.get_user_cache_dir(), "thumbnails")
        return os.path.join(thumb_dir, "normal", thumbnail_hash + ".png"),\
            os.path.join(thumb_dir, "large", thumbnail_hash + ".png")

    @classmethod
    def get_thumbnails_from_xdg_cache(cls, real_uri):
        """Gets pixbufs for the specified thumbnail from the user's cache dir.

        Looks for thumbnails according to the [Thumbnail Managing Standard](https://specifications.freedesktop.org/thumbnail-spec/thumbnail-spec-latest.html#DIRECTORY).

        Args:
            real_uri (str): The URI of the asset.

        Returns:
            List[GdkPixbuf.Pixbuf]: The small thumbnail and the large thumbnail,
            if available in the user's cache directory, otherwise (None, None).
        """
        path_128, path_256 = cls.get_asset_thumbnails_path(real_uri)
        interpolation = GdkPixbuf.InterpType.BILINEAR

        # The cache dirs might have resolutions of 256 and/or 128,
        # while we need 128 (for iconview) and 64 (for listview).
        # First, try the 128 version since that's the native resolution we want.
        try:
            large_thumb = GdkPixbuf.Pixbuf.new_from_file(path_128)
            w, h = large_thumb.get_width(), large_thumb.get_height()
            small_thumb = large_thumb.scale_simple(w / 2, h / 2, interpolation)
            return small_thumb, large_thumb
        except GLib.GError:
            # path_128 doesn't exist, try the 256 version.
            try:
                thumb_256 = GdkPixbuf.Pixbuf.new_from_file(path_256)
                w, h = thumb_256.get_width(), thumb_256.get_height()
                large_thumb = thumb_256.scale_simple(w / 2, h / 2, interpolation)
                small_thumb = thumb_256.scale_simple(w / 4, h / 4, interpolation)
                return small_thumb, large_thumb
            except GLib.GError:
                return None, None

    @classmethod
    def __get_icons(cls, icon_name):
        if icon_name not in cls.icons_by_name:
            small_icon = cls.__get_icon(icon_name, SMALL_THUMB_WIDTH)
            large_icon = cls.__get_icon(icon_name, LARGE_THUMB_WIDTH)
            cls.icons_by_name[icon_name] = (small_icon, large_icon)
        return cls.icons_by_name[icon_name]

    @classmethod
    def __get_icon(cls, icon_name, size):
        icon_theme = Gtk.IconTheme.get_default()
        try:
            icon = icon_theme.load_icon(icon_name, size, Gtk.IconLookupFlags.FORCE_SIZE)
        except GLib.Error:
            icon = icon_theme.load_icon("dialog-question", size, 0)
        return icon

    def _set_state(self):
        # pylint: disable=attribute-defined-outside-init
        asset = self.__asset
        target = asset.get_proxy_target()
        target_is_valid = target and not target.get_error()
        if self.proxy_manager.is_scaled_proxy(asset) and target_is_valid:
            # The asset is a scaled proxy.
            self.state = self.SCALED
        elif self.proxy_manager.is_hq_proxy(asset) and target_is_valid:
            # The asset is a HQ proxy.
            self.state = self.PROXIED
        elif not self.proxy_manager.is_proxy_asset(asset) and asset.proxying_error:
            self.state = self.ASSET_PROXYING_ERROR
        elif self.proxy_manager.is_asset_queued(asset):
            self.state = self.IN_PROGRESS
        elif not asset.is_image() and not self.proxy_manager.is_asset_format_well_supported(asset):
            self.state = self.UNSUPPORTED
        else:
            self.state = self.NO_PROXY

    def decorate(self):
        self._set_state()
        if self.state == self.NO_PROXY:
            self.small_thumb = self.src_small
            self.large_thumb = self.src_large
            return

        self.small_thumb = self.src_small.copy()
        self.large_thumb = self.src_large.copy()

        for thumb in [self.small_thumb, self.large_thumb]:
            emblem = self.EMBLEMS[self.state]
            if thumb.get_height() < emblem.get_height() or \
                    thumb.get_width() < emblem.get_width():
                width = min(emblem.get_width(), thumb.get_width())
                height = min(emblem.get_height(), thumb.get_height())
                # Crop the emblem to fit the thumbnail.
                emblem = emblem.new_subpixbuf(0, emblem.get_height() - height,
                                              width, height)

            # The dest_* arguments define the area of thumb to change.
            # The offset_* arguments define the emblem offset so its
            # bottom-left corner matches the thumb's bottom-left corner.
            emblem.composite(thumb,
                             dest_x=0,
                             dest_y=thumb.get_height() - emblem.get_height(),
                             dest_width=emblem.get_width(),
                             dest_height=emblem.get_height(),
                             offset_x=0,
                             offset_y=thumb.get_height() - emblem.get_height(),
                             scale_x=1.0, scale_y=1.0,
                             interp_type=GdkPixbuf.InterpType.BILINEAR,
                             overall_alpha=self.DEFAULT_ALPHA)


class MediaLibraryWidget(Gtk.Box, Loggable):
    """Widget for managing assets.

    Attributes:
        app (Pitivi): The app.
    """

    __gsignals__ = {
        'play': (GObject.SignalFlags.RUN_LAST, None,
                 (GObject.TYPE_PYOBJECT,))}

    def __init__(self, app):
        Gtk.Box.__init__(self)
        Loggable.__init__(self)

        self._pending_assets = []

        self.app = app
        self._errors = []
        self._project = None
        self._dragged_paths = []
        self.dragged = False
        self.rubberbanding = False
        self.witnessed_tags = set()
        self.tags_popover = None
        self.new_tag_entry = None
        self.__adding_tag = False
        self.clip_view = ViewType.__members__.get(self.app.settings.last_clip_view, ViewType.ICON)
        self.import_start_time = time.time()
        self._last_imported_uris = set()
        self.__last_proxying_estimate_time = _("Unknown")

        self.set_orientation(Gtk.Orientation.VERTICAL)
        builder = Gtk.Builder()
        builder.add_from_file(os.path.join(get_ui_dir(), "medialibrary.ui"))
        builder.connect_signals(self)
        self._welcome_infobar = builder.get_object("welcome_infobar")
        fix_infobar(self._welcome_infobar)
        self._project_settings_infobar = Gtk.InfoBar()
        self._project_settings_infobar.hide()
        self._project_settings_infobar.set_message_type(Gtk.MessageType.OTHER)
        self._project_settings_infobar.set_show_close_button(True)
        self._project_settings_infobar.add_button(_("Project Settings"), Gtk.ResponseType.OK)
        self._project_settings_infobar.connect("response", self.__project_settings_set_infobar_cb)
        self._project_settings_label = Gtk.Label()
        self._project_settings_label.set_line_wrap(True)
        self._project_settings_label.show()
        content_area = self._project_settings_infobar.get_content_area()
        content_area.add(self._project_settings_label)

        fix_infobar(self._project_settings_infobar)
        self._import_warning_infobar = builder.get_object("warning_infobar")
        fix_infobar(self._import_warning_infobar)
        self._import_warning_infobar.hide()
        self._import_warning_infobar.connect("response", self.__warning_infobar_cb)
        self._warning_label = builder.get_object("warning_label")
        self._view_error_button = builder.get_object("view_error_button")
        toolbar = builder.get_object("medialibrary_toolbar")
        toolbar.get_style_context().add_class(Gtk.STYLE_CLASS_INLINE_TOOLBAR)
        self._import_button = builder.get_object("media_import_button")
        self._listview_button = builder.get_object("media_listview_button")
        self.search_entry = builder.get_object("media_search_entry")
        bottom_toolbar_container = builder.get_object("medialibrary_bottom_toolbar_container")
        bottom_toolbar = builder.get_object("medialibrary_bottom_toolbar")
        bg_color = bottom_toolbar_container.get_style_context().get_background_color(Gtk.StateFlags.NORMAL)
        bottom_toolbar.override_background_color(Gtk.StateFlags.NORMAL, bg_color)
        self._clipprops_button = builder.get_object("media_props_button")
        self.tags_button = builder.get_object("tags_button")

        self.scrollwin = Gtk.ScrolledWindow()
        self.scrollwin.set_policy(
            Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.scrollwin.get_accessible().set_name(
            "media_flowbox_scrollwindow")

        self.store = Gio.ListStore()
        self.store.connect("items-changed", self._store_items_changed_cb)

        self.flowbox = Gtk.FlowBox()
        self.flowbox.set_valign(Gtk.Align.START)
        self.flowbox.bind_model(self.store, self.create_iconview_widget_func)
        # Setting up respective adjustments to allow automatic scrolling
        self.flowbox.set_vadjustment(self.scrollwin.get_vadjustment())
        self.flowbox.set_hadjustment(self.scrollwin.get_hadjustment())
        self.scrollwin.add(self.flowbox)

        self.flowbox.connect("button-press-event", self._flowbox_button_press_event_cb)
        self.flowbox.connect("button-release-event", self._flowbox_button_release_event_cb)
        self.flowbox.set_activate_on_single_click(False)
        self.flowbox.connect("child-activated", self._flowbox_child_activated_cb)
        self.flowbox.set_selection_mode(Gtk.SelectionMode.MULTIPLE)
        self.flowbox.connect("selected-children-changed", self._flowbox_selected_children_changed_cb)

        # The _progressbar that shows up when importing clips
        self._progressbar = Gtk.ProgressBar()
        self._progressbar.set_show_text(True)

        # Connect to project.  We must remove and reset the callbacks when
        # changing project.
        project_manager = self.app.project_manager
        project_manager.connect(
            "new-project-loading", self._new_project_loading_cb)
        project_manager.connect("new-project-loaded", self._new_project_loaded_cb)
        project_manager.connect("new-project-failed", self._new_project_failed_cb)
        project_manager.connect("project-closed", self._project_closed_cb)

        # Drag and Drop
        self.drag_dest_set(Gtk.DestDefaults.DROP | Gtk.DestDefaults.MOTION,
                           [URI_TARGET_ENTRY, FILE_TARGET_ENTRY],
                           Gdk.DragAction.COPY)
        self.drag_dest_add_uri_targets()
        self.connect("drag_data_received", self._drag_data_received_cb)

        self._setup_view_as_drag_and_drop_source()

        actions_group = Gio.SimpleActionGroup()
        self.insert_action_group("medialibrary", actions_group)
        self.app.shortcuts.register_group("medialibrary", _("Media Library"), position=50)

        self.remove_assets_action = Gio.SimpleAction.new("remove-assets", None)
        self.remove_assets_action.connect("activate", self._remove_assets_cb)
        actions_group.add_action(self.remove_assets_action)
        self.app.shortcuts.add("medialibrary.remove-assets", ["<Primary>Delete"],
                               self.remove_assets_action,
                               _("Remove the selected assets"))

        # self.update_assets_tag.

        self.insert_at_end_action = Gio.SimpleAction.new("insert-assets-at-end", None)
        self.insert_at_end_action.connect("activate", self._insert_end_cb)
        actions_group.add_action(self.insert_at_end_action)
        self.app.shortcuts.add("medialibrary.insert-assets-at-end", ["Insert"],
                               self.insert_at_end_action,
                               _("Insert selected assets at the end of the timeline"))

        self._update_actions()

        # Set the state of the view mode toggle button.
        self._listview_button.set_active(self.clip_view == ViewType.LIST)
        self.scrollwin.show_all()

        # Add all the child widgets.
        self.pack_start(toolbar, False, False, 0)
        self.pack_start(self._welcome_infobar, False, False, 0)
        self.pack_start(self._project_settings_infobar, False, False, 0)
        self.pack_start(self._import_warning_infobar, False, False, 0)
        self.pack_start(self.scrollwin, True, True, 0)
        self.pack_start(self._progressbar, False, False, 0)
        self.pack_start(bottom_toolbar_container, False, False, 0)

    def create_iconview_widget_func(self, item):
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        box.set_halign(Gtk.Align.CENTER)
        box.props.margin = PADDING / 2

        icon = Gtk.Image.new_from_pixbuf(item.icon_128)

        box.pack_start(icon, False, False, 0)
        box.set_tooltip_markup(item.infotext)
        box.show_all()

        return box

    def create_listview_widget_func(self, item):
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        box.props.margin = PADDING
        box.set_spacing(SPACING)

        icon = Gtk.Image.new_from_pixbuf(item.icon_64)

        label = Gtk.Label()
        label.props.ellipsize = Pango.EllipsizeMode.START
        label.set_markup(item.infotext)
        label.set_yalign(0)

        box.pack_start(icon, False, False, 0)
        box.pack_start(label, False, False, 0)
        box.show_all()

        return box

    def finalize(self):
        self.debug("Finalizing %s", self)

        self.app.project_manager.disconnect_by_func(self._new_project_loading_cb)
        self.app.project_manager.disconnect_by_func(self._new_project_loaded_cb)
        self.app.project_manager.disconnect_by_func(self._new_project_failed_cb)
        self.app.project_manager.disconnect_by_func(self._project_closed_cb)

        if not self._project:
            self.debug("No project set...")
            return

        for asset in self._project.list_assets(GES.Extractable):
            disconnect_all_by_func(asset, self.__asset_proxied_cb)
            disconnect_all_by_func(asset, self.__asset_proxying_cb)

        self.__disconnect_from_project()

    @staticmethod
    def compare_basename_func(model, iter1, iter2, user_data):
        """Compares two model elements."""
        uri1 = model[iter1].uri
        uri2 = model[iter2].uri
        basename1 = GLib.path_get_basename(uri1).lower()
        basename2 = GLib.path_get_basename(uri2).lower()
        if basename1 < basename2:
            return -1
        if basename1 == basename2:
            if uri1 < uri2:
                return -1
        return 1

    def _setup_view_as_drag_and_drop_source(self):
        self.flowbox.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, [URI_TARGET_ENTRY], Gdk.DragAction.COPY)
        self.flowbox.drag_source_add_uri_targets()
        self.flowbox.connect("drag-data-get", self._flowbox_drag_data_get_cb)
        self.flowbox.connect_after("drag-begin", self._flowbox_drag_begin_cb)
        self.flowbox.connect("drag-end", self._flowbox_drag_end_cb)

    def _store_items_changed_cb(self, store_model, unused_position, unused_removed, unused_added):
        if store_model.get_n_items() == 0:
            self._welcome_infobar.show_all()
        else:
            self._welcome_infobar.hide()

    def _import_sources_cb(self, unused_action):
        self.show_import_assets_dialog()

    def _remove_assets_cb(self, unused_action, unused_parameter):
        """Removes the selected assets from the project."""
        assets = self.get_selected_assets()

        with self.app.action_log.started("assets-removal", toplevel=True):
            for asset in assets:
                target = asset.get_proxy_target()
                self._project.remove_asset(asset)
                self.app.gui.editor.timeline_ui.purge_asset(asset.props.id)

                if target:
                    self._project.remove_asset(target)
                    self.app.gui.editor.timeline_ui.purge_asset(target.props.id)

        # The treeview can make some of the remaining items selected, so
        # make sure none are selected.
        self.flowbox.unselect_all()

    def _insert_end_cb(self, unused_action, unused_parameter):
        self.app.gui.editor.timeline_ui.insert_assets(self.get_selected_assets(), -1)

    def _search_entry_changed_cb(self, unused_entry):
        self.filter_store()

    def _search_entry_icon_press_cb(self, entry, icon_pos, event):
        if icon_pos == Gtk.EntryIconPosition.SECONDARY:
            entry.set_text("")
        elif icon_pos == Gtk.EntryIconPosition.PRIMARY:
            self._select_unused_sources()
            # Focus the container so the user can use Ctrl+Delete, for example.
            self.flowbox.grab_focus()

    def filter_store(self):
        text = self.search_entry.get_text().lower()
        # With many hundred clips in an iconview with dynamic columns and
        # ellipsizing, doing needless searches is very expensive.
        # Realistically, nobody expects to search for only one character,
        # and skipping that makes a huge difference in responsiveness.
        if len(text) == 1:
            self.flowbox.show_all()
            return

        # We must convert to markup form to be able to search for &, ', etc.
        text = GLib.markup_escape_text(text)
        for i, row_widget in enumerate(self.flowbox):
            matches = text in self.store[i].infotext.lower()
            row_widget.set_visible(matches)

    def _connect_to_project(self, project):
        """Connects signal handlers to the specified project."""
        project.connect("asset-added", self._asset_added_cb)
        project.connect("asset-loading-progress", self._asset_loading_progress_cb)
        project.connect("asset-removed", self._asset_removed_cb)
        project.connect("error-loading-asset", self._error_creating_asset_cb)
        project.connect("proxying-error", self._proxying_error_cb)
        project.connect("settings-set-from-imported-asset", self.__project_settings_set_from_imported_asset_cb)

    def show_import_assets_dialog(self):
        """Pops up the "Import Sources" dialog box."""
        dialog = Gtk.FileChooserDialog()
        dialog.set_title(_("Select One or More Files"))
        dialog.set_action(Gtk.FileChooserAction.OPEN)
        dialog.set_icon_name("pitivi")
        dialog.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                           _("Add"), Gtk.ResponseType.OK)
        dialog.props.extra_widget = FileChooserExtraWidget(self.app)
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.set_select_multiple(True)
        dialog.set_modal(True)
        dialog.set_transient_for(self.app.gui)
        dialog.set_current_folder(self.app.settings.lastImportFolder)
        dialog.connect('response', self._import_dialog_box_response_cb)
        previewer = PreviewWidget(self.app.settings)
        dialog.set_preview_widget(previewer)
        dialog.set_use_preview_label(False)
        dialog.connect('update-preview', previewer.update_preview_cb)

        # Filter for the "known good" formats by default
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("Supported file formats"))
        file_filter.add_custom(Gtk.FileFilterFlags.URI | Gtk.FileFilterFlags.MIME_TYPE,
                               filter_unsupported_media_files)
        for formatter in GES.list_assets(GES.Formatter):
            for extension in formatter.get_meta("extension").split(","):
                if not extension:
                    continue
                file_filter.add_pattern("*.%s" % extension)
        dialog.add_filter(file_filter)

        # ...and allow the user to override our whitelists
        default = Gtk.FileFilter()
        default.set_name(_("All files"))
        default.add_pattern("*")
        dialog.add_filter(default)

        # Add a shortcut for the project folder (if saved)
        if self._project.uri:
            shortcut = os.path.dirname(self._project.uri)
            dialog.add_shortcut_folder_uri(shortcut)

        dialog.show()

    def _add_asset(self, asset):
        if self.app.proxy_manager.is_proxy_asset(asset) and \
                not asset.props.proxy_target:
            self.info("%s is a proxy asset but has no target, "
                      "not displaying it.", asset.props.id)
            return

        self.debug("Adding asset %s", asset.props.id)

        self._pending_assets.append(asset)

        if self._project.loaded:
            self._flush_pending_assets()

    def update_asset_thumbs(self, asset_uris):
        for item in self.store:
            if item.asset.props.id in asset_uris:
                item.thumb_decorator.disregard_previewer()

    def _flush_pending_assets(self):
        self.debug("Flushing %d pending model rows", len(self._pending_assets))
        for asset in self._pending_assets:
            thumb_decorator = AssetThumbnail(asset, self.app.proxy_manager)
            item = AssetStoreItem(asset, thumb_decorator)

            asset.connect("notify-meta", self.asset_meta_changed_cb)
            self.witnessed_tags.update(item.tags)
            self.store.append(item)

            thumb_decorator.connect("thumb-updated", self.__thumb_updated_cb, asset)

        del self._pending_assets[:]

    def asset_meta_changed_cb(self, asset, meta_key, meta_value):
        if meta_key != "pitivi::tags":
            return

        tags = set(meta_value.split(",")) if meta_value != "" else set()
        for item in self.store:
            if item.asset == asset:
                item.tags = tags
                # Rebuild witnessed_tags
                self.witnessed_tags = set()
                for item_ in self.store:
                    self.witnessed_tags.update(item_.tags)
                break

    def __thumb_updated_cb(self, asset_thumbnail, asset):
        """Handles the thumb-updated signal of the AssetThumbnails in the model."""
        pos = -1
        for i, item in enumerate(self.store):
            if asset == item.asset:
                pos = i
                break

        if pos == -1:
            return

        self.store[pos].icon_64 = asset_thumbnail.small_thumb
        self.store[pos].icon_128 = asset_thumbnail.large_thumb
        selected = self.flowbox.get_child_at_index(pos).is_selected()
        self.store.items_changed(pos, 1, 1)
        if selected:
            self.flowbox.select_child(self.flowbox.get_child_at_index(pos))

    # medialibrary callbacks

    def _asset_loading_progress_cb(self, project, progress, estimated_time):
        self._progressbar.set_fraction(progress / 100)

        proxying_files = []
        for item in self.store:
            item.infotext = beautify_asset(item.asset)
            if not item.asset.ready:
                proxying_files.append(item.asset)
                if item.thumb_decorator.state != AssetThumbnail.IN_PROGRESS:
                    item.thumb_decorator.refresh()
                    item.icon_64 = item.thumb_decorator.small_thumb
                    item.icon_128 = item.thumb_decorator.large_thumb

        if progress == 0:
            self._start_importing()
            return

        if project.loaded:
            if estimated_time:
                self.__last_proxying_estimate_time = beautify_eta(int(
                    estimated_time * Gst.SECOND))

            # Translators: this string indicates the estimated time
            # remaining until an action (such as rendering) completes.
            # The "%s" is an already-localized human-readable duration,
            # such as "31 seconds", "1 minute" or "1 hours, 14 minutes".
            # In some languages, "About %s left" can be expressed roughly as
            # "There remains approximatively %s" (to handle gender and plurals)
            template = ngettext("Transcoding %d asset: %d%% (About %s left)",
                                "Transcoding %d assets: %d%% (About %s left)",
                                len(proxying_files))
            progress_message = template % (
                len(proxying_files), progress,
                self.__last_proxying_estimate_time)
            self._progressbar.set_text(progress_message)

        if progress == 100:
            self._done_importing()

    def __asset_proxying_cb(self, proxy, unused_pspec):
        if not self.app.proxy_manager.is_proxy_asset(proxy):
            self.info("Proxy is not a proxy in our terms (handling deleted proxy"
                      " files while loading a project?) - ignore it")

            return

        self.debug("Proxy is %s - %s", proxy.props.id,
                   proxy.get_proxy_target())
        self.__remove_asset(proxy)

        if proxy.get_proxy_target() is not None:
            # Re add the proxy so its emblem icon is updated.
            self._add_asset(proxy)

    def __asset_proxied_cb(self, asset, unused_pspec):
        self.debug("Asset proxied: %s -- %s", asset, asset.props.id)
        proxy = asset.props.proxy
        self.__remove_asset(asset)
        if not proxy:
            self._add_asset(asset)

        if self._project.loaded:
            self.app.gui.editor.timeline_ui.update_clips_asset(asset)

    def _asset_added_cb(self, unused_project, asset):
        """Checks whether the asset added to the project should be shown."""
        self._last_imported_uris.add(asset.props.id)

        for item in self.store:
            if asset == item.asset:
                self.info("Asset %s already in!", asset.props.id)
                return

        if isinstance(asset, GES.UriClipAsset) and not asset.error:
            self.debug("Asset %s added: %s", asset, asset.props.id)
            asset.connect("notify::proxy", self.__asset_proxied_cb)
            asset.connect("notify::proxy-target", self.__asset_proxying_cb)
            if asset.get_proxy():
                self.debug("Not adding asset %s, its proxy is used instead: %s",
                           asset.props.id,
                           asset.get_proxy().props.id)
                return

            self._add_asset(asset)

    def _asset_removed_cb(self, unused_project, asset):
        if isinstance(asset, GES.UriClipAsset):
            self.debug("Disconnecting %s - %s", asset, asset.props.id)
            asset.disconnect_by_func(self.__asset_proxied_cb)
            asset.disconnect_by_func(self.__asset_proxying_cb)
            self.__remove_asset(asset)

    def __remove_asset(self, asset):
        """Removes the specified asset."""
        uri = asset.get_id()
        # Find the corresponding line in the storemodel and remove it.
        found = False
        for i, item in enumerate(self.store):
            if uri == item.uri:
                self.store.remove(i)
                found = True
                break

        if not found:
            self.info("Failed to remove %s as it was not found"
                      "in the liststore", uri)

    def _proxying_error_cb(self, unused_project, asset):
        self.__remove_asset(asset)
        self._add_asset(asset)

    def _error_creating_asset_cb(self, unused_project, error, asset_id, extractable_type):
        """Gathers asset loading errors."""
        if GObject.type_is_a(extractable_type, GES.UriClip):
            if self.app.proxy_manager.is_proxy_asset(asset_id):
                self.debug("Error %s with a proxy"
                           ", not showing the error message", error)
                return

            self._errors.append((asset_id, str(error.domain), error))

    def _start_importing(self):
        self.__last_proxying_estimate_time = _("Unknown")
        self.import_start_time = time.time()
        self._welcome_infobar.hide()
        self._progressbar.show()

    def _done_importing(self):
        self.debug("Importing took %.3f seconds",
                   time.time() - self.import_start_time)
        self._flush_pending_assets()
        self._progressbar.hide()
        if self._errors:
            errors_amount = len(self._errors)
            btn_text = ngettext("View error", "View errors", errors_amount)
            # Translators: {0:d} is just like %d (integer number variable)
            text = ngettext("An error occurred while importing.",
                            "{0:d} errors occurred while importing.",
                            errors_amount)
            # Do the {0:d} (aka "%d") substitution using "format" instead of %,
            # avoiding tracebacks as %d would be missing in the singular form:
            text = text.format(errors_amount)

            self._view_error_button.set_label(btn_text)
            self._warning_label.set_text(text)
            self._import_warning_infobar.show_all()

        self._select_last_imported_uris()

    def __project_settings_set_from_imported_asset_cb(self, unused_project, asset):
        asset_path = path_from_uri(asset.get_id())
        file_name = os.path.basename(asset_path)
        message = _("The project settings have been set to match file '%s'") % file_name
        self._project_settings_label.set_text(message)
        self._project_settings_infobar.show()

    def _select_last_imported_uris(self):
        if not self._last_imported_uris:
            return
        self._select_sources(self._last_imported_uris)
        self._last_imported_uris = set()

    # Error Dialog Box callbacks

    def _error_dialog_box_close_cb(self, dialog):
        dialog.destroy()

    def _error_dialog_box_response_cb(self, dialog, unused_response):
        dialog.destroy()

    # Import Sources Dialog Box callbacks

    def _import_dialog_box_response_cb(self, dialogbox, response):
        self.debug("response: %r", response)
        if response == Gtk.ResponseType.OK:
            lastfolder = dialogbox.get_current_folder()
            # get_current_folder() is None if file was chosen from 'Recents'
            if not lastfolder:
                lastfolder = GLib.path_get_dirname(dialogbox.get_filename())
            self.app.settings.lastImportFolder = lastfolder
            dialogbox.props.extra_widget.save_values()
            filenames = dialogbox.get_uris()
            self._project.add_uris(filenames)
            if self.app.settings.closeImportDialog:
                dialogbox.destroy()
        else:
            dialogbox.destroy()

    def _source_is_used(self, asset):
        """Checks whether the specified asset is present in the timeline."""
        layers = self._project.ges_timeline.get_layers()
        for layer in layers:
            for clip in layer.get_clips():
                if clip.get_asset() == asset:
                    return True
        return False

    def _select_unused_sources(self):
        """Selects the assets not used by any clip in the project's timeline."""
        unused_sources_uris = []
        for asset in self._project.list_assets(GES.UriClip):
            if not self._source_is_used(asset):
                unused_sources_uris.append(asset.get_id())
        self._select_sources(unused_sources_uris)

    def _select_sources(self, sources_uris):
        self.flowbox.unselect_all()
        for i, item in enumerate(self.store):
            if item.uri in sources_uris:
                self.flowbox.select_child(self.flowbox.get_child_at_index(i))

    # UI callbacks

    def __project_settings_set_infobar_cb(self, infobar, response_id):
        if response_id == Gtk.ResponseType.OK:
            self.app.gui.editor.show_project_settings_dialog()
        infobar.hide()

    def _clip_properties_cb(self, unused_widget):
        """Shows the clip properties in a dialog.

        Allows selecting and applying them as the new project settings.
        """
        assets = self.get_selected_assets()
        if not assets:
            self.debug("No item selected")
            return
        # Only use the first item.
        asset = assets[0]
        dialog = ClipMediaPropsDialog(self._project, asset)
        dialog.dialog.set_transient_for(self.app.gui)
        dialog.run()

    def __warning_infobar_cb(self, infobar, response_id):
        if response_id == Gtk.ResponseType.OK:
            self.__show_errors()
        self._reset_error_list()
        infobar.hide()

    def _reset_error_list(self):
        self._errors = []
        self._import_warning_infobar.hide()

    def __show_errors(self):
        """Shows a dialog with the import errors."""
        title = ngettext("Error while analyzing a file",
                         "Error while analyzing files",
                         len(self._errors))
        headline = ngettext("The following file can not be used with Pitivi.",
                            "The following files can not be used with Pitivi.",
                            len(self._errors))
        error_dialogbox = FileListErrorDialog(title, headline)
        error_dialogbox.connect("close", self._error_dialog_box_close_cb)
        error_dialogbox.connect("response", self._error_dialog_box_response_cb)

        for uri, reason, extra in self._errors:
            error_dialogbox.add_failed_file(uri, reason, extra)
        error_dialogbox.window.set_transient_for(self.app.gui)
        error_dialogbox.window.show()

    def _toggle_view_type_cb(self, widget):
        paths = self.get_selected_paths()
        if widget.get_active():
            self.clip_view = ViewType.LIST
            func = self.create_listview_widget_func
        else:
            self.clip_view = ViewType.ICON
            func = self.create_iconview_widget_func
        self.app.settings.last_clip_view = self.clip_view.name
        self.flowbox.bind_model(self.store, func)
        for path in paths:
            child = self.flowbox.get_child_at_index(path)
            self.flowbox.select_child(child)
        # In case the toggling is done when the items are filtered.
        self.filter_store()

    def _tags_button_clicked_cb(self, unused_widget):
        self.tags_popover = Gtk.Popover()
        popover_heading = Gtk.Label(label=_("Tag as:"))
        popover_heading.set_halign(Gtk.Align.START)

        paths = self.get_selected_paths()

        # Find the common tags in the selected assets
        common_tags = list(self.store[paths[0]].tags)
        for path in paths:
            for tag in common_tags:
                if tag not in self.store[path].tags:
                    common_tags.remove(tag)
            if not common_tags:
                break

        # Union of tags associated with assets under the current selection.
        all_tags = []
        for path in paths:
            all_tags = list(set().union(all_tags, self.store[path].tags))
        inconsistent_tags = list(set(all_tags) - set(common_tags))

        tags = list(self.witnessed_tags)
        tags.sort()
        tagstore = Gio.ListStore()
        for tag in tags:
            if tag in common_tags:
                state = TagState.PRESENT
            elif tag in inconsistent_tags:
                state = TagState.INCONSISTENT
            else:
                state = TagState.ABSENT
            tagstore.append(TagStoreItem(tag, state))

        tags_list = Gtk.ListBox()
        tags_list.bind_model(tagstore, self.create_tagslist_widget_func)
        tags_list.set_can_focus(False)

        new_entry_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.new_tag_entry = Gtk.Entry()
        self.new_tag_entry.props.placeholder_text = _("Enter tag")
        self.new_tag_entry.set_width_chars(12)
        self.new_tag_entry.connect("activate", self.new_tag_entry_activated_cb, tagstore)
        add_tag_button = Gtk.Button.new_with_label(_("Add"))
        add_tag_button.connect("clicked", self.add_tag_button_clicked_cb, tagstore)
        new_entry_box.pack_start(self.new_tag_entry, False, False, 0)
        new_entry_box.pack_start(add_tag_button, False, False, PADDING)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.props.margin = PADDING
        box.pack_start(popover_heading, False, False, SPACING)
        box.pack_start(tags_list, True, True, PADDING)
        box.pack_start(new_entry_box, False, False, 0)

        self.tags_popover.connect("closed", self._tags_popover_closed_cb, tags_list, tagstore)
        self.tags_popover.add(box)
        self.tags_popover.set_position(Gtk.PositionType.BOTTOM)
        self.tags_popover.set_relative_to(self.tags_button)
        self.tags_popover.show_all()
        self.tags_popover.popup()
        tags_list.unselect_all()

    def create_tagslist_widget_func(self, tag_item: TagStoreItem):
        """Converts a tag item to a widget."""
        box = Gtk.ListBoxRow()
        box.props.margin = PADDING

        checkbutton = Gtk.CheckButton.new_with_label(tag_item.name)

        if tag_item.initial_state == TagState.INCONSISTENT:
            checkbutton.props.inconsistent = True
        active = self.__adding_tag or tag_item.initial_state == TagState.PRESENT
        checkbutton.set_active(active)

        checkbutton.connect("toggled", self._tag_toggled_cb, tag_item)

        box.add(checkbutton)
        box.show_all()

        return box

    def _tags_popover_closed_cb(self, popover, tags_list, tagstore):
        self.apply_changed_tags(tags_list, tagstore)
        popover.hide()
        self.tags_button.set_active(False)
        self.tags_popover = None

    def _tag_toggled_cb(self, toggle_button, tag_item):
        """Handles the toggling of a tag in the list."""
        if toggle_button.props.inconsistent:
            toggle_button.props.active = True
            toggle_button.props.inconsistent = False
        else:
            if toggle_button.props.active:
                if tag_item.initial_state == TagState.INCONSISTENT:
                    toggle_button.props.inconsistent = True

    def new_tag_entry_activated_cb(self, unused_widget, tagstore):
        self._add_new_tag(tagstore)

    def add_tag_button_clicked_cb(self, unused_widget, tagstore):
        self._add_new_tag(tagstore)

    def _add_new_tag(self, tagstore):
        if not self.new_tag_entry.get_text():
            return

        tag_name = self.new_tag_entry.get_text()
        # Don't accept duplicate entry
        for tag_item in tagstore:
            if tag_item.name == tag_name:
                return

        self.__adding_tag = True
        try:
            tagstore.append(TagStoreItem(tag_name, TagState.ABSENT))
        finally:
            self.__adding_tag = False

        self.new_tag_entry.set_text("")

    def apply_changed_tags(self, tags_list, tagstore):
        paths = self.get_selected_paths()
        with self.app.action_log.started("Alter tags", toplevel=True):
            for i, row_widget in enumerate(tags_list):
                checkbox = row_widget.get_child()
                initial_state = tagstore[i].initial_state
                tag_name = tagstore[i].name
                if checkbox.props.inconsistent:
                    # Nothing shall be changed.
                    continue

                if checkbox.props.active:
                    if initial_state != TagState.PRESENT:
                        for path in paths:
                            asset_store_item = self.store[path]
                            if tag_name not in asset_store_item.tags:
                                old_meta = asset_store_item.asset.get_meta("pitivi::tags")
                                new_meta = old_meta + "," + tag_name if old_meta else tag_name
                                asset_store_item.asset.set_meta("pitivi::tags", new_meta)
                else:
                    if initial_state != TagState.ABSENT:
                        for path in paths:
                            asset_store_item = self.store[path]
                            if tag_name in asset_store_item.tags:
                                old_meta = asset_store_item.asset.get_meta("pitivi::tags")
                                new_meta = old_meta.split(",")
                                new_meta.remove(tag_name)
                                new_meta = ",".join(new_meta)
                                asset_store_item.asset.set_meta("pitivi::tags", new_meta)

    def __stop_using_proxy_cb(self, unused_action, unused_parameter):
        prefer_original = self.app.settings.proxying_strategy == ProxyingStrategy.NOTHING
        self._project.disable_proxies_for_assets(self.get_selected_assets(),
                                                 hq_proxy=not prefer_original)

    def __use_proxies_cb(self, unused_action, unused_parameter):
        self._project.use_proxies_for_assets(self.get_selected_assets())

    def __use_scaled_proxies_cb(self, unused_action, unused_parameter):
        self._project.use_proxies_for_assets(self.get_selected_assets(),
                                             scaled=True)

    def __delete_proxies_cb(self, unused_action, unused_parameter):
        prefer_original = self.app.settings.proxying_strategy == ProxyingStrategy.NOTHING
        self._project.disable_proxies_for_assets(self.get_selected_assets(),
                                                 delete_proxy_file=True,
                                                 hq_proxy=not prefer_original)

    def __open_containing_folder_cb(self, unused_action, unused_parameter):
        assets = self.get_selected_assets()
        if len(assets) != 1:
            return
        parent_path = os.path.dirname(path_from_uri(assets[0].get_id()))
        Gio.AppInfo.launch_default_for_uri(Gst.filename_to_uri(parent_path), None)

    def __edit_nested_clip_cb(self, unused_action, unused_parameter):
        assets = self.get_selected_assets()
        if len(assets) != 1:
            return

        path = os.path.abspath(path_from_uri(assets[0].get_id()))
        subprocess.Popen([sys.argv[0], path])

    def __create_menu_model(self):
        if self.app.proxy_manager.proxying_unsupported:
            return None, None

        assets = self.get_selected_assets()
        if not assets:
            return None, None

        action_group = Gio.SimpleActionGroup()
        menu_model = Gio.Menu()

        if len(assets) == 1:
            action = Gio.SimpleAction.new("open-folder", None)
            action.connect("activate", self.__open_containing_folder_cb)
            action_group.insert(action)
            text = _("Open containing folder")
            menu_model.append(text, "assets.%s" % action.get_name().replace(" ", "."))

        if len(assets) == 1 and assets[0].props.is_nested_timeline:
            action = Gio.SimpleAction.new("edit-nested-clip", None)
            action.connect("activate", self.__edit_nested_clip_cb)
            action_group.insert(action)
            text = _("Edit")
            menu_model.append(text, "assets.%s" % action.get_name().replace(" ", "."))

        image_assets = [asset for asset in assets
                        if asset.is_image()]

        if len(assets) == len(image_assets):
            return menu_model, action_group

        video_streams = []
        for asset in assets:
            video_streams += [
                stream_info
                for stream_info in asset.get_info().get_stream_list()
                if isinstance(stream_info, GstPbutils.DiscovererVideoInfo)]

        proxies = [asset.get_proxy_target() for asset in assets
                   if self.app.proxy_manager.is_proxy_asset(asset)]
        hq_proxies = [asset.get_proxy_target() for asset in assets
                      if self.app.proxy_manager.is_hq_proxy(asset)]
        scaled_proxies = [asset.get_proxy_target() for asset in assets
                          if self.app.proxy_manager.is_scaled_proxy(asset)]
        in_progress = [asset.creation_progress for asset in assets
                       if asset.creation_progress < 100]

        if hq_proxies:
            action = Gio.SimpleAction.new("unproxy-asset", None)
            action.connect("activate", self.__stop_using_proxy_cb)
            action_group.insert(action)
            text = ngettext("Do not use Optimised Proxy for selected asset",
                            "Do not use Optimised Proxies for selected assets",
                            len(proxies) + len(in_progress))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

            action = Gio.SimpleAction.new("delete-proxies", None)
            action.connect("activate", self.__delete_proxies_cb)
            action_group.insert(action)

            text = ngettext("Delete corresponding proxy file",
                            "Delete corresponding proxy files",
                            len(proxies) + len(in_progress))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

        if in_progress:
            action = Gio.SimpleAction.new("unproxy-asset", None)
            action.connect("activate", self.__stop_using_proxy_cb)
            action_group.insert(action)
            text = ngettext("Do not use Proxy for selected asset",
                            "Do not use Proxies for selected assets",
                            len(proxies) + len(in_progress))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

            action = Gio.SimpleAction.new("delete-proxies", None)
            action.connect("activate", self.__delete_proxies_cb)
            action_group.insert(action)

            text = ngettext("Delete corresponding proxy file",
                            "Delete corresponding proxy files",
                            len(proxies) + len(in_progress))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

        if scaled_proxies:
            action = Gio.SimpleAction.new("unproxy-asset", None)
            action.connect("activate", self.__stop_using_proxy_cb)
            action_group.insert(action)
            text = ngettext("Do not use Scaled Proxy for selected asset",
                            "Do not use Scaled Proxies for selected assets",
                            len(proxies) + len(in_progress))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

            action = Gio.SimpleAction.new("delete-proxies", None)
            action.connect("activate", self.__delete_proxies_cb)
            action_group.insert(action)

            text = ngettext("Delete corresponding proxy file",
                            "Delete corresponding proxy files",
                            len(proxies) + len(in_progress))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

        if len(proxies) != len(assets) and len(in_progress) != len(assets):
            action = Gio.SimpleAction.new("use-proxies", None)
            action.connect("activate", self.__use_proxies_cb)
            action_group.insert(action)
            text = ngettext("Use Optimised Proxy for selected asset",
                            "Use Optimised Proxies for selected assets", len(assets))

            menu_model.append(text, "assets.%s" %
                              action.get_name().replace(" ", "."))

            if video_streams:
                action = Gio.SimpleAction.new("use-scaled-proxies", None)
                action.connect("activate", self.__use_scaled_proxies_cb)
                action_group.insert(action)
                text = ngettext("Use Scaled Proxy for selected asset",
                                "Use Scaled Proxies for selected assets", len(assets))

                menu_model.append(text, "assets.%s" %
                                  action.get_name().replace(" ", "."))

        return menu_model, action_group

    def _flowbox_selected_children_changed_cb(self, flowbox):
        self._update_actions()

    def _update_actions(self):
        selected_count = len(self.flowbox.get_selected_children())
        self.remove_assets_action.set_enabled(selected_count)
        self.insert_at_end_action.set_enabled(selected_count)
        self.tags_button.set_sensitive(selected_count)
        # Some actions can only be done on a single item at a time:
        self._clipprops_button.set_sensitive(selected_count == 1)

    def _flowbox_child_activated_cb(self, unused_flowbox, child):
        """Plays the asset identified by the specified path.

        This can happen when an item is double-clicked, or
        Space, Return or Enter is pressed.
        """
        path = child.get_index()
        self.emit("play", self.store[path].asset)

    def _flowbox_button_press_event_cb(self, flowbox, event):
        child = flowbox.get_child_at_pos(event.x, event.y)
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
            # It is possible to double-click outside of clips
            pass
        else:
            if not event.get_state() & (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK):
                # Ctrl or Shift key is not pressed.
                # It is a left or right mouse click.
                if child:
                    # Reset the selection to only select the clicked asset.
                    self.rubberbanding = False
                    if not child.is_selected():
                        self.flowbox.unselect_all()
                        child.grab_focus()
                        self.flowbox.select_child(child)
                else:
                    # The click was done over an empty space.
                    # Clear the selection.
                    self.rubberbanding = True
                    self.flowbox.unselect_all()
            else:
                # Either Ctrl or Shift key is pressed.
                self.rubberbanding = False

        # Returning True propagates the signal further enabling the default
        # behaviour of rubberband selection. Returning False stops the
        # propagation to enable drag&drop of assets from the Media Library
        # to the Timeline.
        return self.rubberbanding

    def _flowbox_button_release_event_cb(self, flowbox, event):
        if self.rubberbanding:
            self.rubberbanding = False

        res, button = event.get_button()
        if not res or button != 3:
            return

        child = self.flowbox.get_child_at_pos(event.x, event.y)
        if child:
            if not child.is_selected():
                self.flowbox.unselect_all()
                self.flowbox.select_child(child)

        model, action_group = self.__create_menu_model()
        if not model or not model.get_n_items():
            self.debug("Not showing popup menu")
            return

        popover = Gtk.Popover.new_from_model(self.flowbox, model)
        popover.insert_action_group("assets", action_group)
        popover.props.position = Gtk.PositionType.BOTTOM

        rect = Gdk.Rectangle()
        rect.x = event.x
        rect.y = event.y
        rect.width = 1
        rect.height = 1
        popover.set_pointing_to(rect)
        popover.show_all()

    def __disconnect_from_project(self):
        self._project.disconnect_by_func(self._asset_added_cb)
        self._project.disconnect_by_func(self._asset_loading_progress_cb)
        self._project.disconnect_by_func(self._asset_removed_cb)
        self._project.disconnect_by_func(self._proxying_error_cb)
        self._project.disconnect_by_func(self._error_creating_asset_cb)
        self._project.disconnect_by_func(self.__project_settings_set_from_imported_asset_cb)

    def _new_project_loading_cb(self, project_manager, project):
        assert not self._project

        self._project = project
        self._reset_error_list()
        self.store.remove_all()
        self._welcome_infobar.show_all()
        self._connect_to_project(project)

    def _new_project_loaded_cb(self, project_manager, project):
        # Make sure that the sources added to the project are added
        self._flush_pending_assets()

    def _new_project_failed_cb(self, project_manager, uri, reason):
        self.store.remove_all()
        self._project = None

    def _project_closed_cb(self, project_manager, project):
        self.__disconnect_from_project()
        self._project_settings_infobar.hide()
        self.store.remove_all()
        self._project = None

    def __paths_walked_cb(self, uris):
        """Handles the end of the path walking when importing dragged dirs."""
        if not uris:
            return

        if not self._project:
            return

        # At the end of the import operation, these will be selected.
        self._last_imported_uris = set(uris)
        assets = self._project.assets_for_uris(uris)
        if assets:
            # All the files have already been added.
            # This is the only chance we have to select them.
            self._select_last_imported_uris()
        else:
            self._project.add_uris(uris)

    def _drag_data_received_cb(self, widget, context, x, y,
                               selection, targettype, time_):
        """Handles data being dragged onto self."""
        self.debug("targettype: %d, selection.data: %r",
                   targettype, selection.get_data())
        uris = selection.get_uris()
        # Scan in the background what was dragged and
        # import whatever can be imported.
        self.app.threads.add_thread(PathWalker, uris, self.__paths_walked_cb)

    def _flowbox_drag_data_get_cb(self, view, context, data, info, timestamp):
        uris = [self.store[path].uri for path in self._dragged_paths]
        data.set_uris(uris)

    def _flowbox_drag_begin_cb(self, view, context):
        self.dragged = True
        self._dragged_paths = self.get_selected_paths()

        if not self._dragged_paths:
            context.drag_abort(int(time.time()))
        else:
            row = self.store[self._dragged_paths[0]]
            icon = row.icon_128

            icon_height = icon.get_height()
            icon_width = icon.get_width()

            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, icon_width, icon_height)
            ctx = cairo.Context(surface)
            # Center the icon around the cursor.
            ctx.translate(icon_width / 2, icon_height / 2)
            surface.set_device_offset(-icon_width / 2, -icon_height / 2)

            Gdk.cairo_set_source_pixbuf(ctx, icon, 0, 0)
            ctx.paint_with_alpha(0.35)

            Gtk.drag_set_icon_surface(context, surface)

    def _flowbox_drag_end_cb(self, view, context):
        self.info("Drag operation ended")
        self.dragged = False

    def get_selected_paths(self):
        selected_children = self.flowbox.get_selected_children()
        paths = []
        for child in selected_children:
            paths.append(child.get_index())
        return paths

    def get_selected_assets(self):
        """Gets the selected assets."""
        paths = self.get_selected_paths()
        return [self.store[path].asset
                for path in paths]

    def activate_compact_mode(self):
        self._import_button.set_is_important(False)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the mediafilespreviewer module."""
# pylint: disable=attribute-defined-outside-init,protected-access
from unittest import mock

from gi.repository import GLib

from pitivi.dialogs.missingasset import MissingAssetDialog
from pitivi.mediafilespreviewer import PreviewWidget
from pitivi.utils.proxy import ProxyingStrategy
from tests import common


class PreviewWidgetTest(common.TestCase):
    """Tests for the PreviewWidget class."""

    def test_select_missing_asset(self):
        """Exercise the MissingAssetDialog when loading a project."""
        app = common.create_pitivi(proxying_strategy=ProxyingStrategy.NOTHING,
                                   FCpreviewWidth=640,
                                   FCpreviewHeight=480)

        proj_uri = self.create_project_file_from_xges("""<ges version='0.3'>
            <project properties='properties;' metadatas='metadatas;'>
                <ressources>
                    <asset id='file://this/is/a/moved/asset.mp4' extractable-type-name='GESUriClip'
                        properties='properties, supported-formats=(int)6, duration=(guint64)1228000000;' metadatas='metadatas' />
                </ressources>
            </project>
            </ges>""")
        project_manager = app.project_manager

        # Use a cloned sample so the asset is not in GES's asset cache.
        # This combination of calls can lead to a mainloop freeze:
        # - MissingAssetDialog.get_new_uri() through the "missing-uri" signal handler,
        # - MissingAssetDialog.run() through MissingAssetDialog.get_new_uri(),
        # - GES.UriClipAsset.request_sync() through PreviewWidget.preview_uri,
        with common.cloned_sample("1sec_simpsons_trailer.mp4"):
            asset_uri = common.get_sample_uri("1sec_simpsons_trailer.mp4")

            mainloop = common.create_main_loop()

            def missing_uri_cb(project_manager, project, unused_error, asset):
                with mock.patch.object(MissingAssetDialog, "set_transient_for"):
                    mad = MissingAssetDialog(app, asset, asset.get_id())
                mad._chooser.select_uri(asset_uri)
                # Close the dialog when idle so get_new_uri does not get stuck.
                GLib.idle_add(mad.close)
                uri = mad.get_new_uri()
                return uri

            project_manager.connect("missing-uri", missing_uri_cb)

            preview_loaded_for_uri = ""

            def preview_uri(preview_widget, uri):
                nonlocal preview_loaded_for_uri
                original_preview_uri(preview_widget, uri)
                # If it gets past the original_preview_uri call, it's all fine!
                preview_loaded_for_uri = uri
                mainloop.quit()

            original_preview_uri = PreviewWidget.preview_uri
            PreviewWidget.preview_uri = preview_uri
            try:
                # Our mainloop timeout mechanism cannot be used,
                # because the mainloop gets blocked.
                with common.CheckedOperationDuration(seconds=2):
                    project_manager.load_project(proj_uri)
                    mainloop.run()
                self.assertEqual(preview_loaded_for_uri, asset_uri)
            finally:
                PreviewWidget.preview_uri = original_preview_uri



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Millan Castro <m.castrovilarino@gmail.com>
# Copyright (c) 2021, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the timeline.markers module."""
from unittest import mock

from gi.repository import Gdk
from gi.repository import Gtk

from pitivi.utils.timeline import Zoomable
from tests import common


class TestMarkers(common.TestCase):
    """Tests for markers."""

    @common.setup_timeline
    def test_marker_added_ui(self):
        """Checks the add marker UI."""
        markers = self.timeline.get_marker_list("markers")
        marker_box = self.timeline_container.markers

        x = 100
        event = mock.Mock(spec=Gdk.EventButton)
        event.x = x
        event.y = 1
        event.button = Gdk.BUTTON_PRIMARY

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = marker_box
            event.guiEvent = Gdk.Event.new(Gdk.EventType.BUTTON_PRESS)
            marker_box.do_button_press_event(event)
            event.guiEvent = Gdk.Event.new(Gdk.EventType.BUTTON_RELEASE)
            marker_box.do_button_release_event(event)

        position = Zoomable.pixel_to_ns(event.x)
        self.assert_markers(markers, [(position, None)])

    @common.setup_timeline
    def test_marker_removed_ui(self):
        """Checks the remove marker UI."""
        markers = self.timeline.get_marker_list("markers")
        marker_box = self.timeline_container.markers

        x = 200
        position = Zoomable.pixel_to_ns(x)
        marker = marker_box.markers_container.add(position)
        self.assert_markers(markers, [(position, None)])

        event = mock.Mock(spec=Gdk.EventButton)
        event.x = x
        event.y = 1
        event.button = Gdk.BUTTON_SECONDARY

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = marker.ui
            event.guiEvent = Gdk.Event.new(Gdk.EventType.BUTTON_PRESS)
            marker_box.do_button_press_event(event)
            event.guiEvent = Gdk.Event.new(Gdk.EventType.BUTTON_RELEASE)
            marker_box.do_button_release_event(event)

        self.assert_markers(markers, [])

    @common.setup_timeline
    def test_marker_moved_ui(self):
        """Checks the move marker UI."""
        markers = self.timeline.get_marker_list("markers")
        marker_box = self.timeline_container.markers

        x1 = 300
        position1 = Zoomable.pixel_to_ns(x1)
        marker = marker_box.markers_container.add(position1)
        self.assert_markers(markers, [(position1, None)])

        x2 = 400
        position2 = Zoomable.pixel_to_ns(x2)

        event = mock.Mock(spec=Gdk.EventButton)
        event.x = x2
        event.y = 1
        event.type = Gdk.EventType.BUTTON_PRESS
        event.button = Gdk.BUTTON_PRIMARY

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = marker.ui
            event.guiEvent = Gdk.Event.new(Gdk.EventType.BUTTON_PRESS)
            marker_box.do_button_press_event(event)

            with mock.patch.object(marker.ui, "translate_coordinates") as translate_coordinates:
                translate_coordinates.return_value = (x2, 0)
                marker_box.do_motion_notify_event(event)
                marker_box.do_button_release_event(event)

        self.assert_markers(markers, [(position2, None)])

    @common.setup_timeline
    def test_marker_comment_ui(self):
        """Checks the comments marker UI."""
        markers = self.timeline.get_marker_list("markers")
        marker_box = self.timeline_container.markers

        x = 500
        position = Zoomable.pixel_to_ns(x)
        marker = marker_box.markers_container.add(position)
        self.assert_markers(markers, [(position, None)])

        event = mock.Mock(spec=Gdk.EventButton)
        event.x = x
        event.y = 1
        event.type = Gdk.EventType.BUTTON_PRESS
        event.button = Gdk.BUTTON_PRIMARY

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = marker.ui

            def popup(markerpopover):
                text_buffer = markerpopover.text_view.get_buffer()
                text_buffer.set_text("com")
                text_buffer.set_text("comment")
                markerpopover.popdown()
            original_popover_menu = Gtk.Popover.popup
            Gtk.Popover.popup = popup
            try:
                event.guiEvent = Gdk.Event.new(Gdk.EventType.BUTTON_PRESS)
                marker_box.do_button_press_event(event)
                event.guiEvent = Gdk.Event.new(Gdk.EventType.DOUBLE_BUTTON_PRESS)
                event.type = Gdk.EventType.DOUBLE_BUTTON_PRESS
                marker_box.do_button_press_event(event)
            finally:
                Gtk.Popover.popup = original_popover_menu

        stack, = self.action_log.undo_stacks
        self.assertEqual(len(stack.done_actions), 1, stack.done_actions)

        self.assert_markers(markers, [(position, "comment")])

    def check_seek(self, action, current_position, expected_position):
        pipeline = self.project.pipeline
        with mock.patch.object(pipeline, "get_position") as get_position:
            get_position.return_value = current_position
            with mock.patch.object(pipeline, "simple_seek") as simple_seek:
                action.activate()
                if expected_position is None:
                    simple_seek.assert_not_called()
                else:
                    simple_seek.assert_called_once_with(expected_position)

    @common.setup_timeline
    def test_seeking_actions(self):
        """Checks the seeking actions."""
        # The seek logic ignores the markers past the timeline duration
        # since it's not possible to seek there.
        self.add_clip(self.timeline.layers[0], start=0, duration=20)

        markers = self.timeline.get_marker_list("markers")
        marker_box = self.timeline_container.markers

        marker_box.markers_container.add(10)
        marker_box.markers_container.add(12)
        self.assert_markers(markers, [(10, None), (12, None)])

        self.check_seek(marker_box.seek_forward_marker_action, 9, 10)
        self.check_seek(marker_box.seek_forward_marker_action, 10, 12)
        self.check_seek(marker_box.seek_forward_marker_action, 11, 12)
        self.check_seek(marker_box.seek_forward_marker_action, 12, None)
        self.check_seek(marker_box.seek_forward_marker_action, 13, None)

        self.check_seek(marker_box.seek_backward_marker_action, 9, None)
        self.check_seek(marker_box.seek_backward_marker_action, 10, None)
        self.check_seek(marker_box.seek_backward_marker_action, 11, 10)
        self.check_seek(marker_box.seek_backward_marker_action, 12, 10)
        self.check_seek(marker_box.seek_backward_marker_action, 13, 12)
